import itertools
from tqdm import tqdm
import borg
import numpy as np


def build_gravity_model(state, box, cpar):
    print(cpar)
    lpt = borg.forward.models.BorgLpt(box, box, ai=1.0)
    chain = borg.buildDefaultChain(box, cpar, 1.0, lpt)
    return chain, lpt


def build_likelihood(state, info):
    return borg.likelihood.GaussianLinear(info)


if not borg.EMBEDDED and __name__ == "__main__":
    from tqdm import tqdm

    Ng =  8

    cpar = borg.cosmo.CosmologicalParameters()
    state = borg.likelihood.MarkovState()
    info = borg.likelihood.LikelihoodInfo()
    info["GRID_LENGTH"] = np.array([0, 100., 0, 100., 0, 100.])
    info["GRID"] = np.array([Ng, Ng, Ng], dtype=np.uint32)
    box = borg.forward.BoxModel(100., Ng)

    state.newArray3d("galaxy_data_0", Ng, Ng, Ng)

    fwd, lpt = build_gravity_model(state, box, cpar)
    state.newForwardModel("BORG_model", fwd)
    state.newScalar("corner0", 0.0)
    state.newScalar("corner1", 0.0)
    state.newScalar("corner2", 0.0)
    state.newScalar("localNdata0", 0, False, "L")
    state.newScalar("localNdata1", Ng, False, "L")
    state.newScalar("localNdata2", 0, False, "L")
    state.newScalar("localNdata3", Ng, False, "L")
    state.newScalar("localNdata4", 0, False, "L")
    state.newScalar("localNdata5", Ng, False, "L")
    state.newScalar("ares_heat", 1.0)
    state.newScalar("NCAT", 1, False, "L")
    state.newScalar("galaxy_bias_ref_0", False)

    state.newScalar("cosmology", cpar)
    state.newScalar("galaxy_nmean_0", 0.0)
    state.newArray1d("galaxy_bias_0", 2)
    state.newArray3d("galaxy_synthetic_sel_window_0", Ng, Ng, Ng)
    state["galaxy_bias_0"][:] = np.array([1.0, 1.0])
    state["galaxy_synthetic_sel_window_0"][:] = 1

    like = build_likelihood(state, info)

    like.initializeLikelihood(state)
    like.updateMetaParameters(state)

    s_hat = np.fft.rfftn(np.random.randn(Ng, Ng, Ng) / Ng**(1.5))
    like.generateMockData(s_hat, state)

    like.logLikelihood(s_hat)
    analytic_gradient = like.gradientLikelihood(s_hat)

    num_gradient = np.zeros((Ng, Ng, Ng // 2 + 1), dtype=np.complex128)
    s_hat_epsilon = s_hat.copy()

    epsilon = 0.001
    for i, j, k in tqdm(itertools.product(*map(range, [Ng, Ng, Ng // 2 + 1])),
                        total=Ng * Ng * (Ng // 2 + 1)):
        s_hat_epsilon[i, j, k] = s_hat[i, j, k] + epsilon
        L = like.logLikelihood(s_hat_epsilon)
        s_hat_epsilon[i, j, k] = s_hat[i, j, k] - epsilon
        L -= like.logLikelihood(s_hat_epsilon)
        QQ = L / (2.0 * epsilon)

        s_hat_epsilon[i, j, k] = s_hat[i, j, k] + 1j * epsilon
        L = like.logLikelihood(s_hat_epsilon)
        s_hat_epsilon[i, j, k] = s_hat[i, j, k] - 1j * epsilon
        L -= like.logLikelihood(s_hat_epsilon)
        QQ = QQ + L * 1j / (2.0 * epsilon)

        s_hat_epsilon[i, j, k] = s_hat[i, j, k]

        num_gradient[i, j, k] = QQ

    np.savez("gradients.npz", num=num_gradient, ana=analytic_gradient)



import h5py as h5
import cosmotool as ct
import numpy as np
import ares_tools as at


#result = at.read_chain_avg_dev(".",["final_density", "v_field"],step=20*4,prefix="output",do_dev=True,start=20000)
if False:
  mean_p,dev_p = at.read_chain_complex_avg_dev(".",lambda ff: (1+ff['final_density'][...])*ff['v_field'][...],step=10,prefix="output",do_dev=True,start=200, pattern="%s_%04d.h5")
  mean_d,dev_d= at.read_chain_complex_avg_dev(".",lambda ff: ff['final_density'][...],step=10,prefix="output",do_dev=True,start=200,pattern="%s_%04d.h5")
  np.savez("means.npz", mean_p=mean_p, dev_p=dev_p, mean_d=mean_d,dev_d=dev_d)
else:
  x=np.load("means.npz")
  mean_d = x['mean_d']
  mean_p = x['mean_p']

with h5.File("output_2200.h5", mode="r") as ff:
  one_d = ff['final_density'][...]
  one_p = ff['v_field'][...]#(ff['final_density'][...]+1)*ff['v_field'][...]

Nside=256
L=4000
N=256
Dmin=10
Dmax=128

#ix=np.arange(N)*L/N - 0.5*L
ix=np.arange(N)*L/N

# These are the box corners
corner_x=-2200
corner_y=-2000
corner_z=-300

shifter=-np.array([corner_x,corner_y,corner_z])*N/L - 0.5*N
shifter2=np.array([corner_x,corner_y,corner_z])
mm=ct.spherical_projection(Nside, mean_d, Dmin, Dmax,shifter=shifter,integrator_id=1,booster=100)
one_mm=ct.spherical_projection(Nside, one_d, Dmin, Dmax,shifter=shifter,integrator_id=1,booster=100)

x = ix[:,None,None].repeat(N,axis=1).repeat(N,axis=2) + shifter2[0]
y = ix[None,:,None].repeat(N,axis=0).repeat(N,axis=2) + shifter2[1]
z = ix[None,None,:].repeat(N,axis=0).repeat(N,axis=1) + shifter2[2]
r = np.sqrt(x**2+y**2+z**2)

cond = r>0
pr = np.where(cond, (mean_p[0,...] * x + mean_p[1,...] * y + mean_p[2,...]*z)/r,0)
one_pr = np.where(cond,(one_p[0,...] * x + one_p[1,...] * y + one_p[2,...]*z)/r,0)

one = np.ones(pr.shape)

mpr=ct.spherical_projection(Nside, pr, Dmin, Dmax,shifter=shifter,integrator_id=1,booster=100)
one_mpr=ct.spherical_projection(Nside, one_pr, Dmin, Dmax,shifter=shifter,integrator_id=1,booster=100)
mdist=ct.spherical_projection(Nside, one, Dmin, Dmax,shifter=shifter,integrator_id=1,booster=100)

np.savez("sky.npz", d=mm,pr=mpr,r=r,dist=mdist,one_d=one_mm,one_pr=one_mpr)





#+
#   ARES/HADES/BORG Package -- ./extra/python/python/aquila_borg/__init__.py
#   Copyright (C) 2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
"""
PyBORG root module
==================

PyBORG is a complete binding of the borg machine to python. The borg
module can be imported either from a normal python interpret or from
the embedded python in hades_python.

Be aware that you cannot import directly any of the native modules. To import
all borg facilities, it is required *and* sufficient to just execute ``import aquila_borg``.
It will provide all the modules such as `aquila_borg.cosmo` or `aquila_borg.samplers`.


*Note:* The `borg` module may be installed as an alias to `aquila_borg`. However the package name
is the same as the one used by BorgBackup. Thus there may be clashes between these two softwares.
It is advised to use `aquila_borg` to avoid conflicts.

"""
import sys
import io
import contextlib
import logging

try:
    import mpi4py

    MPI4PY_AVAILABLE = True
except:
    MPI4PY_AVAILABLE = False

from .borg_native import *

from . import dft
from . import utils

class _RewriteIO(io.IOBase):
    def __init__(self, original):
        self.original = original
        self.newline = False
        if MPI4PY_AVAILABLE:
            self.comm = mpi4py.MPI.COMM_WORLD
            self.prefix = f"[{self.comm.rank} / {self.comm.size}] "
        else:
            self.prefix = "[0/0]"

    def write(self, b):
        if b[0] == "\n":
            return self.original.write(b)
        return self.original.write(self.prefix + b)

    def writelines(self, lines):
        return self.original.writelines(lines)

class BorgMessageStream(io.IOBase):
    def __init__(self, console, level='std'):
        super().__init__()
        self.console = console
        if level == 'std':
            self.printer = console.print_std
        elif level == 'warn':
            self.printer = console.print_warning
        elif level == 'error':
            self.printer = console.print_error
        else:
            raise ValueError()

    def write(self, b):
        for line in b.split(sep="\n"):
           self.printer(line)

    def writelines(self, lines):
        for line in lines:
            if line[-1] == '\n':
                line = line[:-1]
            self.printer(line)

def capture_warnings(logger="py.warnings",level='warn'):
    logging.captureWarnings(True)
    l = logging.getLogger(logger)
    h = logging.StreamHandler(BorgMessageStream(console(), level=level))
    l.addHandler(h)
    return h

@contextlib.contextmanager
def borgConsole(original=sys.stdout):
    """Setup a clean redirection of BORG output, notably in presence of MPI4PY.

    Keyword Arguments:
        original (file-like): Where to send the console to (default: :class:`sys.stdout`)

    """
    yield contextlib.redirect_stdout(_RewriteIO(original))


def buildDefaultChain(box, cosmo_pars, a_init, model):
    """Build a default BORG chain from a model. This remove the hastle of setting up
    a primordial fluctuation generator, and the cosmological parameters.

    Args:
        box (borg.forward.BoxModel): the specification of the box for the simulation
        cosmo_pars (borg.cosmo.CosmologicalParameters): the cosmology chosen
        a_init (float): the starting scale factor for `model`
        model (borg.forward.BORGForwardModel): the model to be chained with

    Returns:
        borg.forward.BORGForwardModel: a newly built chain for cosmology
    """
    chain = forward.ChainForwardModel(box)
    chain.addModel(forward.models.HermiticEnforcer(box))
    chain.addModel(forward.models.Primordial(box, a_init))
    chain.addModel(forward.models.EisensteinHu(box))
    chain.addModel(model)
    chain.setCosmoParams(cosmo_pars)
    return chain


if EMBEDDED:
    capture_warnings()

__all__ = ["borgConsole", "buildDefaultChain", "dft", "utils"] + borg_native.__all__

# ARES TAG: authors_num = 1
# ARES TAG: name(0) = Guilhem Lavaux
# ARES TAG: year(0) = 2020
# ARES TAG: email(0) = guilhem.lavaux@iap.fr



#+
#   ARES/HADES/BORG Package -- ./extra/hades/scripts/generate_benchmark_tests.py
#   Copyright (C) 2019 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+

import os
import sys

output_cmake_file = sys.argv[1]
input_config = sys.argv[2]
base_dir = sys.argv[3]
input_config_path = os.path.dirname(os.path.abspath(input_config))

with open(input_config, mode="r") as f:
  config = {}
  exec(f.read(), {}, config)
  config = config['tests']

tests = config['tests']

with open(output_cmake_file, "wt") as f:

    for test_name in tests:
      test = tests[test_name]
      includes = test['includes']
      likelihood = test['likelihood']
      info_code = test.get('info_code', '')
      downgrade = test.get('downgrade', 1)

      model_code=""

      f.write("""
add_executable(benchmark_gradient_%(test_name)s %(base_dir)s/%(test_source)s)
target_link_libraries(benchmark_gradient_%(test_name)s test_library_LSS LSS ${LIBS})
ares_add_test_targets(benchmark_gradient_%(test_name)s)
target_include_directories(benchmark_gradient_%(test_name)s PRIVATE %(input_config_path)s)
"""
         % dict(test_name=test_name, base_dir=base_dir, input_config_path=input_config_path, test_source="benchmark_%s.cpp" % (test_name,)
      ))

      includes_str = \
        "\n".join(
            map(lambda x: "#include \"%s\"" % (x,), includes)
        )

      if 'model' in test:
        if type(test['model']) is list:
             model_string = ""
             args = test.get(
                 'model_args', ['comm, box,box, 0.001'] + ['comm, box, box, 1']*(len(test['model'])-1))
             extra_code = test.get(
                 'model_extra', [""]*len(test['model'])
             )
             extra_code_prev = test.get(
                 'model_extra_prev', [""]*len(test['model'])
             )
             for m, a, e, eprev in zip(test['model'], args, extra_code, extra_code_prev):
                 model_string += f"""
                                  {{
                                    {eprev}
                                    auto m = std::make_shared<{m}>({a}); 
                                    {e}
                                    chain->addModel(m);
                                  }}"""
                 model_code = f"""
                                auto buildModel(LibLSS::MPI_Communication * comm, LibLSS::MarkovState& state, LibLSS::BoxModel box, LibLSS::BoxModel box2) {{
                                  using namespace LibLSS;
                                  auto chain = std::make_shared<ChainForwardModel>(comm, box);
                                  {model_string}
                                  return chain;
                                }}
                 """
        else:
          args = test.get('model_args', 'comm, box, box, 0.001')
          model_code="auto buildModel(LibLSS::MPI_Communication *comm, LibLSS::MarkovState& state, LibLSS::BoxModel const& box, LibLSS::BoxModel box2) { return std::make_shared<%(model)s>(%(model_args)s); }" % dict(model=test['model'],model_args=args )


      if 'data_setup' in test:
          data_setup='#include "%s"\n#define DATA_SETUP %s\n'%(test['data_setup'])
      else:
          data_setup=''

      with open(os.path.join(base_dir,"benchmark_%s.cpp" % (test_name,)), mode="wt") as f2:
        f2.write("""%(includes)s
#include "libLSS/physics/likelihoods/base.hpp"
%(data_setup_header)s

static const int DOWNGRADE_DATA = %(downgrade)d;

namespace L = LibLSS::Likelihood;
using LibLSS::LikelihoodInfo;

auto makeLikelihood(LikelihoodInfo& info) {
  %(additional_info_code)s
  return std::make_shared<%(likelihood)s>(info);
}

%(model_code)s

static constexpr bool RUN_RSD_TEST = false;
static std::string testName = "%(test_name)s";

#include "generic_gradient_benchmark.cpp"
""" % dict(includes=includes_str,data_setup_header=data_setup, additional_info_code=info_code,likelihood=likelihood,model_code=model_code,test_name=test_name,downgrade=downgrade))

# ARES TAG: authors_num = 1
# ARES TAG: name(0) = Guilhem Lavaux
# ARES TAG: email(0) = guilhem.lavaux@iap.fr
# ARES TAG: year(0) = 2019



#cython: language_level=3
import numpy as np
cimport numpy as np

np.import_array()
np.import_ufunc()

cdef extern from "sys/types.h":
   ctypedef np.int64_t int64_t

cdef extern from "numpy/npy_common.h":
   ctypedef npy_intp

cdef extern from "special_math.hpp" namespace "CosmoTool":
   T log_modified_bessel_first_kind[T](T v, T z) nogil except +

cdef extern from "numpy_adaptors.hpp" namespace "CosmoTool":
   void parallel_ufunc_dd_d[T,IT](char **args, IT* dimensions, IT* steps, void *func)


cdef np.PyUFuncGenericFunction loop_func[1]
cdef char input_output_types[3]
cdef void *elementwise_funcs[1]

loop_func[0] = <np.PyUFuncGenericFunction>parallel_ufunc_dd_d[double,npy_intp]

input_output_types[0] = np.NPY_DOUBLE
input_output_types[1] = np.NPY_DOUBLE
input_output_types[2] = np.NPY_DOUBLE

elementwise_funcs[0] = <void*>log_modified_bessel_first_kind[double]

log_modified_I = np.PyUFunc_FromFuncAndData(
    loop_func,
    elementwise_funcs,
    input_output_types,
    1, # number of supported input types
    2, # number of input args
    1, # number of output args
    0, # `identity` element, never mind this
    "log_modified_bessel_first_kind", # function name
    "log_modified_bessel_first_kind(v: Float, z: Float) -> Float", # docstring
    0 # unused
    )




#+
#   ARES/HADES/BORG Package -- ./build_tools/gen_code_in_header.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import sys
import re

code=""
with open(sys.argv[1], mode="r") as f_in, open(sys.argv[2], mode="w") as f_out:
  for line in f_in:
    line = re.sub(r'\\', r'\\\\', line)
    line = re.sub(r'"', r'\"', line)
    line = line[:-1]
    code += line + "\\n"
  f_out.write('"%s"' % (code,))



#+
#   ARES/HADES/BORG Package -- ./scripts/test_analysis.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
from analysis.analysis import *
import pylab as plt
import numpy as np
import healpy as hp

#chain_path="/scratch/jasche/panphasia_run_pm/"
chain_path="/scratch/jasche/2mpp_highres_pm/"

ares=analysis(chain_path=chain_path,LSS_framework='BORG')


mu_i,var_i,mu_f,var_i=ares.mean_var_density(first_sample=299,last_sample=300,sample_steps=2)


plt.imshow(np.log(2+mu_f[:,:,64]))
plt.show()
'''
image=ares.get_spherical_slice(mu,nside=256,rslice=50)
hp.mollview(image)
plt.show()

'''

'''
k,mu,var=ares.mean_var_spec(first_sample=0,last_sample=100000,sample_steps=1)

#set loglog scale
plt.xscale('log')
plt.yscale('log')

plt.errorbar(k, mu, yerr=np.sqrt(var), fmt='-')
#plt.plot(k,mu,color='red')

plt.show()
'''



#+
#   ARES/HADES/BORG Package -- ./scripts/ini_generator/gen_subcat_conf.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import sys
import os.path
import argparse

pp = argparse.ArgumentParser()
pp.add_argument('--output', type=str, required=True)
pp.add_argument('--configs', type=str, required=True)
pp.add_argument('--header', type=str, required=True)
args = pp.parse_args()
out_ini= args.output
all_config_files=args.configs.split(':')



subcat_id=0

PATTERN="""datafile=%(catalog)s
maskdata=%(mask)s
"""

def apply_cut_magnitude(CAT_config, j):
      CUT_PATTERN="""galaxy_bright_absolute_magnitude_cut=%(absmag_bright)15.15lf
galaxy_faint_absolute_magnitude_cut=%(absmag_faint)15.15lf
"""
      Nsubcat = CAT_config['num_subcat']
      DeltaMag = (CAT_config['absmag_max'] - CAT_config['absmag_min'])
      MagMin = CAT_config['absmag_min']
      absmag_bright = DeltaMag * j / Nsubcat + MagMin
      absmag_faint = DeltaMag * (j+1) / Nsubcat + MagMin
      f.write(CUT_PATTERN % {'absmag_bright':absmag_bright,'absmag_faint':absmag_faint})

def apply_cut_distance(CAT_config, j):
      CUT_PATTERN="""file_dmin=%(dmin)15.15lf
file_dmax=%(dmax)15.15lf
"""
      Nsubcat = CAT_config['num_subcat']
      DeltaMag = (CAT_config['d_max'] - CAT_config['d_min'])
      MagMin = CAT_config['d_min']
      dmin = DeltaMag * j / Nsubcat + MagMin
      dmax = DeltaMag * (j+1) / Nsubcat + MagMin
      f.write(CUT_PATTERN % {'dmin':dmin,'dmax':dmax})

def execfile(filename, globals=None, locals=None):
    if globals is None:
        globals = sys._getframe(1).f_globals
    if locals is None:
        locals = sys._getframe(1).f_locals
    with open(filename, "r") as fh:
        exec(fh.read()+"\n", globals, locals)

with open(out_ini, mode="wt") as f:

  with open(args.header, mode="rt") as fh:
    f.write(fh.read())
    f.write("\n")

  print("All configs = %r" % all_config_files)
  for config_file in all_config_files:
    path_config,_ = os.path.split(config_file)
    def file_subs(s):
      return os.path.join(path_config,s)

    config_locals={}
    config_globals={'FILE':file_subs}

    print("Analyze %s" % config_file)
    execfile(config_file, config_globals, config_locals)

    CAT_config = config_locals['CONFIG']
    del config_locals['CONFIG']
    
    CAT_config['catalog'] = os.path.join(path_config,CAT_config['catalog'])
    if CAT_config['cutter']=='magnitude':
      cut_function = apply_cut_magnitude
    elif CAT_config['cutter']=='distance':
      cut_function = apply_cut_distance
    else:
      print("Unknown cutter '%s'" % CAT_config['cutter'])
      sys.exit(1)

    Nsubcat = CAT_config['num_subcat']
    for j in range(Nsubcat):
      f.write("[catalog_%(subcat_id)d]\n" % {'subcat_id':subcat_id})
      for k,v in config_locals.items():
        if type(v)==str:
          f.write("%s=%s\n" % (k,v))
        elif type(v)==tuple:
          if len(v) > 0:
            f.write((("%s=" + "%r,"*len(v)) % ((k,) + v))[:-1] + "\n")
        else:
          f.write("%s=%r\n" % (k,v))
      cut_function(CAT_config, j)
      f.write(PATTERN % CAT_config)   
      if (j==CAT_config.get('ref_subcat',-1)):
        f.write("refbias=true\n")
      else:
        f.write("refbias=false\n")
      f.write("\n")
      
      subcat_id += 1
      
  f.write("[run]\nNCAT=%d\n\n" % subcat_id)



#
# This python script gives an example on using JAX and PyBORG together
#
import jax
import numpy as np
import borg

cons = borg.console()

myprint=lambda x: cons.print_std(x) if type(x) == str else cons.print_std(repr(x))
myprint("Hello!")

@jax.jit
def jax_func(rho):
   return rho**2

class MyModel(borg.forward.BaseForwardModel):
   def __init__(self, box):
     myprint("Start forward model")
     super().__init__(box, box)

   def getPreferredInput(self):
     return borg.forward.PREFERRED_REAL

   def getPreferredOutput(self):
     return borg.forward.PREFERRED_REAL

   def forwardModel_v2_impl(self, input_array):
     # Save the input data in a jax array (i.e. upload to accelerator)
     self.save = jax.numpy.array(input_array)
 
   def getDensityFinal_impl(self, output_array):
     # Run forward, and save the AG function
     fwd, self.ag_fun = jax.vjp(jax_func, self.save)
     output_array[:] = fwd

   def adjointModel_v2_impl(self, input_ag):
     # Save the ag vector
     self.ag = input_ag
   
   def getAdjointModel_impl(self, output_ag):
     # Evaluate the new function with ag_fun
     out_ag, = self.ag_fun(self.ag)
     output_ag[:] = out_ag

def build_gravity_model(box):
  chain = borg.forward.ChainForwardModel(box)
  chain.addModel(borg.forward.models.Primordial(box, 1.0))
  chain.addModel(borg.forward.models.EisensteinHu(box))
  chain.addModel(borg.forward.models.BorgLpt(box, box, ai=1.0))
  chain.addModel(MyModel(box))
  return chain


if __name__ == "__main__":
  box = borg.forward.BoxModel()
  cpar = borg.cosmo.CosmologicalParameters()

  chain = build_gravity_model(box)
  chain.setCosmoParams(cpar)

  s_hat = np.fft.rfftn(np.random.randn(*box.N)/np.sqrt(box.N[0]**3))

  myprint(np.var(s_hat))

  chain.forwardModel_v2(s_hat)
  rho = np.zeros(chain.getOutputBoxModel().N)
  chain.getDensityFinal(rho)


  ag = 2*rho
  chain.adjointModel_v2(ag)
  chain.getAdjointModel(ag)



#+
#   ARES/HADES/BORG Package -- ./libLSS/tests/plot_grav.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import h5py as h5
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

H100=100.e3
h=0.68
L=100.
N=128
G=6.67e-11
omega_m=0.30
Mpc_in_m=3.08567758e22
udistance=1.*Mpc_in_m

dmean = 1.0/(float(N)**3)

with h5.File("gravity.h5") as f:
  g = f['gravity'][...]
  p = f['position'][...]
  pot = f['potential'][...]
  ud = f['unit_density'][0] 
  up = f['unit_potential'][0]

ud *= Mpc_in_m**3

g = g[:(g.shape[0]/2),:]
p = p[:(p.shape[0]/2),:]
pot = pot[:(pot.shape[0]/2)]

ref = np.array([L/2,0,L/2])

plt.clf()
#plt.plot(-g[:,0])
plt.plot(p[:,1],-g[:,1])

yy = p[:,1]

aa = 6.67e-11 * ud * (L/N)**3 * yy/yy**3

plt.plot(yy, aa)

plt.gca().set_yscale('log')
plt.gca().set_xscale('log')

#plt.plot(g[:,2])

plt.gcf().savefig("grav.png")


xx=np.arange(N/2)*L/N 
mass = 3*(H100*h/Mpc_in_m)**2/(8*np.pi*G) * omega_m * (Mpc_in_m)**3

real_pot = 6.67e-11 * mass / (udistance*xx)

plt.clf()
plt.plot(xx,pot*up)
plt.plot(xx,real_pot)
plt.gca().set_yscale('log')
plt.gca().set_xscale('log')
plt.gcf().savefig("pot.png")



import matplotlib.pyplot as plt
import numpy as np
import borg
import jax
import jax.numpy as jnp

myprint = lambda s: borg.console().print_std(s if type(s) == str else repr(s))


def get_psi(k2, k, x, L, delta_hat, inv_dVol):
    phi_hat = jnp.where(k2 == 0.0, 0.0j, delta_hat * 1j * k / (k2+1e-19))
    phi_hat[0,0,0]=0j
    p = x + jnp.fft.irfftn(phi_hat) * inv_dVol
    return jnp.where(p >= 0, p % L, L - (-p) % L)


def mymodf(x):
    ix = jnp.floor(x)
    return x - ix, ix


def do_cic(x, y, z, Nx, Ny, Nz, Lx, Ly, Lz):

    Ntot = Nx * Ny * Nz
    x = x * Nx / Lx
    y = y * Ny / Ly
    z = z * Nz / Lz
    qx, ix = mymodf(x)
    qy, iy = mymodf(y)
    qz, iz = mymodf(z)
    ix = ix.astype(int)
    iy = iy.astype(int)
    iz = iz.astype(int)
    rx = 1.0 - qx
    ry = 1.0 - qy
    rz = 1.0 - qz
    jx = (ix + 1) % Nx
    jy = (iy + 1) % Ny
    jz = (iz + 1) % Nz

    rho = jnp.zeros((Ntot, ))

    for a in [False, True]:
        for b in [False, True]:
            for c in [False, True]:
                ax = jx if a else ix
                ay = jy if b else iy
                az = jz if c else iz
                ux = qx if a else rx
                uy = qy if b else ry
                uz = qz if c else rz

                idx = az + Nz * ay + Nz * Ny * ax
                rho += jnp.bincount(idx, weights=ux * uy * uz, length=Ntot)

    return rho.reshape((Nx, Ny, Nz)) / (x.shape[0] / Ntot) - 1.0


def build_density(delta_hat, ikx, iky, ikz, x, y, z, N0: int, N1: int, N2: int,
                  L0: float, L1: float, L2: float, inv_dVol: float):
    k2 = ikx**2 + iky**2 + ikz**2
    return do_cic(
        get_psi(k2, ikx, x, L0, delta_hat, inv_dVol).flatten(),
        get_psi(k2, iky, y, L1, delta_hat, inv_dVol).flatten(),
        get_psi(k2, ikz, z, L2, delta_hat, inv_dVol).flatten(), N0, N1, N2, L0,
        L1, L2)


build_density = jax.jit(build_density, (7, 8, 9))


class JaxLpt(borg.forward.BaseForwardModel):
    def __init__(self, box):
        super().__init__(box, box)
        N0, N1, N2 = box.N
        L0, L1, L2 = box.L
        self.ikx = 2 * np.pi * jnp.fft.fftfreq(N0, d=L0 / N0)[:, jnp.newaxis,
                                                              jnp.newaxis]
        self.iky = 2 * np.pi * jnp.fft.fftfreq(N1, d=L1 / N1)[jnp.newaxis, :,
                                                              jnp.newaxis]
        self.ikz = 2 * np.pi * jnp.fft.fftfreq(
            N2, d=L2 / N2)[jnp.newaxis, jnp.newaxis, :(N2 // 2 + 1)]
        self.x = jnp.arange(N0)[:, jnp.newaxis, jnp.newaxis] * L0 / N0
        self.y = jnp.arange(N1)[jnp.newaxis, :, jnp.newaxis] * L1 / N1
        self.z = jnp.arange(N2)[jnp.newaxis, jnp.newaxis, :] * L2 / N2
        self.inv_dVol = box.Ntot / box.volume
        self.L = tuple(box.L)
        self.N = tuple(box.N)

    def getPreferredInput(self):
        return borg.forward.PREFERRED_FOURIER

    def getPreferredOutput(self):
        return borg.forward.PREFERRED_REAL

    def forwardModel_v2_impl(self, input_array):
        import jax.numpy as jnp
        myprint("In fwdmodel!")
        delta_hat = jnp.array(input_array)
        self.save_array, self.grad = jax.vjp(
            lambda d: build_density(d, self.ikx, self.iky, self.ikz, self.x,
                                    self.y, self.z, self.N[0], self.N[1], self.
                                    N[2], self.L[0], self.L[1], self.L[
                                        2], self.inv_dVol), delta_hat)

    def getDensityFinal_impl(self, output_array):
        output_array[:] = self.save_array

    def adjointModel_v2_impl(self, input_ag):
        ag_delta = jnp.array(input_ag)
        self.output_ag, = self.grad(ag_delta)

    def getAdjointModel_impl(self, output_ag):
        output_ag[:] = self.output_ag

    def setModelParams(self, myparams):
        print(f"setModelParams in python: pars={myparams}")


cons = borg.console()

cons.setVerboseLevel(5)

cosmo_par = borg.cosmo.CosmologicalParameters()
cosmo_par.default()
print(repr(cosmo_par))

cosmo = borg.cosmo.Cosmology(cosmo_par)

box = borg.forward.BoxModel()
box.L = (100, 100, 100)
box.N = (64,64,64)#, 32, 32)
print(box)

rho = np.zeros(box.N)
rho_gpu = np.zeros(box.N)
icplus = np.zeros((box.N[0], box.N[1], box.N[2] // 2 + 1), dtype=np.complex128)

ic = np.fft.rfftn(np.random.randn(*box.N)) / box.N[0]**(1.5)

box2 = borg.forward.BoxModel()
box2.L = box.L
box2.N = tuple(map(lambda x: 2 * x, box.N))

h = borg.forward.models.HermiticEnforcer(box)
primordial = borg.forward.models.Primordial(box, 1.0)
ehu = borg.forward.models.EisensteinHu(box)

chain = borg.forward.ChainForwardModel(box)
chain.addModel(h)
chain.addModel(primordial)
chain.addModel(ehu)
chain.setCosmoParams(cosmo_par)

#lpt_gpu = borg.forward.models.newModel("LPT_GPU", box, {});
lpt = borg.forward.models.newModel(
    "LPT_CIC", box,
    dict(a_initial=1.0,
         a_final=1.0,
         do_rsd=False,
         supersampling=1,
         lightcone=False,
         part_factor=1.1,
         mul_out=1))
lpt_omp = borg.forward.models.newModel(
    "LPT_CIC_OPENMP", box,
    dict(a_initial=1.0,
         a_final=1.0,
         do_rsd=False,
         supersampling=1,
         lightcone=False,
         part_factor=1.1,
         mul_out=1))

rho = np.zeros(box.N)
rho_gpu = np.zeros(box.N)
rho_omp = np.zeros(box.N)
icplus = np.zeros(box.N)

#chain.addModel(lpt_gpu)
chain.forwardModel_v2(ic)
chain.getDensityFinal(icplus)

lpt_gpu = JaxLpt(box)

#
lpt.setCosmoParams(cosmo_par)
#lpt_omp.setCosmoParams(cosmo_par)
#ehu.getDensityFinal(icplus)
lpt.forwardModel_v2(icplus)
lpt.getDensityFinal(rho)
lpt_gpu.forwardModel_v2(icplus)
lpt_gpu.getDensityFinal(rho_gpu)
#lpt_omp.forwardModel_v2(icplus)
#lpt_omp.getDensityFinal(rho_omp)
#
#
ag = np.random.uniform(size=box.N)/10.
lpt_gpu.adjointModel_v2(ag)
ag_gpu = np.zeros(box.N)
lpt_gpu.getAdjointModel(ag_gpu)

lpt.adjointModel_v2(ag)
ag_cpu = np.zeros(box.N)
lpt.getAdjointModel(ag_cpu)



#+
#   ARES/HADES/BORG Package -- ./scripts/regraft_run.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import warnings

warnings.simplefilter("ignore", FutureWarning)
import h5py as h5
import argparse
import os

class NotCompatible(Exception):
    def __init__(msg_):
        self.msg = msg_

    def __repr__():
        return "NotCompatible: " + self.msg

def restart_name(prefix, cpu):
    return os.path.join(prefix, "restart.h5_%d" % cpu)

def detect_ncpu(prefix):
    ncpu = 0
    while True:
        try:
            os.stat(os.path.join(prefix, "restart.h5_%d" % ncpu))
        except OSError:
            break
        ncpu += 1
    return ncpu

def bias_name(bid):
    return "scalars/galaxy_bias_%d" % bid

def nmean_name(nid):
    return "scalars/galaxy_nmean_%d" % nid

def check_compat(restart_prefix, mcmc):
    print("Checking compatibility of MCMC and restart")
    with h5.File(restart_name(restart_prefix, 0), mode="r") as f_r, \
        h5.File(mcmc, mode="r") as f_m:

        bname = bias_name(0)
        while bname in f_m:
            if not bname in f_r:
                raise NotCompatible("Not enough catalogs")

            if f_m[bname].size != f_r[bname].size:
                raise NotCompatible("Incompatible bias model")

            bias_num += 1
            bname = bias_name(bias_num)

def transfer(src, dest, name, start):
    if (not name in src) or (not name in dest):
        return False

    sz = dest[name].shape[0]
    dest[name][...] = src[name][start:(start+sz),...]
    return True

def checked_transfer(*args):
    if not transfer(*args):
        raise NotCompatible("Problems in grafting")

def graft(restart_prefix, ncpus, mcmc):
    with h5.File(args.mcmc, mode="r") as f_m:
        plane_start = 0
        plane_len = 0

        for cpu in range(ncpus):
            print("Transplanting to restart CPU %d" % cpu)
            with h5.File(restart_name(restart_prefix, cpu), mode="r") as f_r:
                plane_len = f_r['scalars/BORG_final_density'].shape[0]
                checked_transfer(f_m, f_r, 'scalars/BORG_final_density', plane_start)
                checked_transfer(f_m, f_r, 'scalars/s_field', plane_start)
                checked_transfer(f_m, f_r, 'scalars/s_hat_field', plane_start)

                bias_num = 0
                while True:
                    if not transfer(f_m, f_r, bias_name(bias_num), 0):
                        break
                    if not transfer(f_m, f_r, nmean_name(bias_num), 0):
                        break
                    bias_num += 1

                plane_start += plane_len

p = argparse.ArgumentParser(description="Graft the state of a previous run on a given restart.")

p.add_argument('mcmc', type=str, help="MCMC state to import")
p.add_argument('restart', type=str, help="restart prefix directory")
args = p.parse_args()

ncpu = detect_ncpu(args.restart)
print("Found %d CPU restart file")


check_compat(args.restart, args.mcmc)
graft(args.restart, ncpu, args.mcmc)

print("Done")



#+
#   ARES/HADES/BORG Package -- ./extra/borg/scripts/generate_vobs_tests.py
#   Copyright (C) 2019 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+

import os
import sys

output_cmake_file = sys.argv[1]
input_config = sys.argv[2]
base_dir = sys.argv[3]

with open(input_config, mode="r") as f:
  config = {}
  exec(f.read(), {}, config)
  config = config['tests']

tests = config['tests']

with open(output_cmake_file, "wt") as f:

    for test_name in tests:
      test = tests[test_name]
      includes = test['includes']
      likelihood = test['likelihood']
      model = test['model']
      model_args = test.get('model_args', 'comm, box, 0.001');
      model_code=""

      f.write("""
add_executable(test_vobs_%(test_name)s %(base_dir)s/%(test_source)s)
target_link_libraries(test_vobs_%(test_name)s test_library_LSS LSS ${LIBS})
ares_add_test_targets(test_vobs_%(test_name)s)
"""
         % dict(test_name=test_name, base_dir=base_dir, test_source="test_%s_vobs.cpp" % (test_name,)
      ))

      includes_str = \
        "\n".join(
            map(lambda x: "#include \"%s\"" % (x,), includes)
        )


      if 'model' in test:
        args = test.get('model_args', 'comm, box, 0.001')

      with open(os.path.join(base_dir,"test_%s_vobs.cpp" % (test_name,)), mode="wt") as f2:
        f2.write("""%(includes)s
#include "libLSS/samplers/rgen/hmc/hmc_density_sampler.hpp"
#include "libLSS/physics/likelihoods/base.hpp"

namespace L = LibLSS::Likelihood;
using LibLSS::LikelihoodInfo;
using LibLSS::HMCDensitySampler;
using LibLSS::MarkovState;
using LibLSS::BoxModel;

typedef %(likelihood)s ThisLikelihood;


auto makeModel(MarkovState& state, BoxModel const& box, LikelihoodInfo& info) {
  auto comm = L::getMPI(info);
  return std::make_shared<%(model)s>(%(model_args)s);
}

#include "libLSS/tests/generic_borg_vobs_test.cpp"
""" % dict(includes=includes_str,likelihood=likelihood,model=model,model_args=model_args))

# ARES TAG: authors_num = 1
# ARES TAG: name(0) = Guilhem Lavaux
# ARES TAG: email(0) = guilhem.lavaux@iap.fr
# ARES TAG: year(0) = 2019



#+
#   ARES/HADES/BORG Package -- ./scripts/lic_plot/lic_demo.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import numpy as np
import pylab as plt

import lic_internal

dpi = 100
size = 700
video = False

vortex_spacing = 0.5
extra_factor = 2.

a = np.array([1,0])*vortex_spacing
b = np.array([np.cos(np.pi/3),np.sin(np.pi/3)])*vortex_spacing
rnv = int(2*extra_factor/vortex_spacing)
vortices = [n*a+m*b for n in range(-rnv,rnv) for m in range(-rnv,rnv)]
vortices = [(x,y) for (x,y) in vortices if -extra_factor<x<extra_factor and -extra_factor<y<extra_factor]


xs = np.linspace(-1,1,size).astype(np.float32)[None,:]
ys = np.linspace(-1,1,size).astype(np.float32)[:,None]

vectors = np.zeros((size,size,2),dtype=np.float32)
for (x,y) in vortices:
    rsq = (xs-x)**2+(ys-y)**2
    vectors[...,0] +=  (ys-y)/rsq
    vectors[...,1] += -(xs-x)/rsq
    
texture = np.random.rand(size,size).astype(np.float32)

plt.bone()
frame=0

if video:
    kernellen = 31
    for t in np.linspace(0,1,16*5):
        kernel = np.sin(np.arange(kernellen)*np.pi/kernellen)*(1+np.sin(2*np.pi*5*(np.arange(kernellen)/float(kernellen)+t)))

        kernel = kernel.astype(np.float32)

        image = lic_internal.line_integral_convolution(vectors, texture, kernel)

        plt.clf()
        plt.axis('off')
        plt.figimage(image)
        plt.gcf().set_size_inches((size/float(dpi),size/float(dpi)))
        plt.savefig("flow-%04d.png"%frame,dpi=dpi)
        frame += 1
else:
    kernellen=31
    kernel = np.sin(np.arange(kernellen)*np.pi/kernellen)
    kernel = kernel.astype(np.float32)

    image = lic_internal.line_integral_convolution(vectors, texture, kernel)

    plt.clf()
    plt.axis('off')
    plt.figimage(image)
    plt.gcf().set_size_inches((size/float(dpi),size/float(dpi)))
    plt.savefig("flow-image.png",dpi=dpi)





#+
#   ARES/HADES/BORG Package -- ./scripts/misc/check_integrator.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import h5py as h5
import numpy as np
import pylab as plt


fig = plt.figure(1)

fig.clf()
ax = fig.add_subplot(111)

f = h5.File("symplectic.h5")
for k in f.keys():
  ax.semilogy(np.abs(f[k]['energy']), label=k)
f.close()

ax.legend()
fig.savefig("symplectic.png")



#+
#   ARES/HADES/BORG Package -- ./scripts/old_analysis/analysis.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
from read_all_h5 import explore_chain
from read_all_h5 import rebuild_spliced_h5
import h5py as h5
import numpy as np
import healpy as hp
import numexpr as ne
import os
import math
from pylab import *
import matplotlib.pyplot as plt
from matplotlib.pyplot import figure, show

#import numba
'''
@numba.jit
def _special_spectrum_builder(P, PP, tmp, alpha, N_i, Pk_i):

    for i in range(Pk_i.size):
        for j in range(PP.size):
            v = PP[j]**(-alpha)*exp(- N_i[i]*Pk_i[i] / (2 * PP[j]))
            tmp[j] = v
            total += v

        for j in range(PP.size):
            P[j] += tmp[j] / total
 ''' 
 
 
 #ARES/HADES/BORG image scanning class
class IndexTracker:
    def __init__(self, ax, X):
        self.ax = ax
        self.ax.set_title('use up/down keys to navigate images')

        self.X = X
        rows,cols,self.slices = X.shape
        self.ind  = self.slices/2

        cmax=X.max()
        cmin=X.min()
        
        self.im = self.ax.imshow(self.X[:,self.ind,:],vmax=cmax,vmin=cmin)
        self.update()

    def onscroll(self, event):
        #print ("%s " % (event.key))
        if event.key=='up':
            self.ind = np.clip(self.ind+1, 0, self.slices-1)
        else:
            self.ind = np.clip(self.ind-1, 0, self.slices-1)


        self.update()

    def update(self):
        self.im.set_data(self.X[:,self.ind,:])
        self.ax.set_ylabel('slice %s'%self.ind)
        self.im.axes.figure.canvas.draw()

           
#ARES/HADES/BORG analysis class
class analysis:

    def __init__(self, chain_path='.', LSS_framework='ARES'):
        self.chain_path 	= chain_path
        self.LSS_framework 	= LSS_framework
      
        self.description = "This Class is part of the ARES/HADES/BORG analysis framework"
        self.author = "Copyright (C) 2009-2016 Jens Jasche \n Copyright (C) 2014-2016 Guilhem Lavaux"
        
        #get chain setup
        
        self.L0=0
        self.L1=0
        self.L2=0
        self.N0=0
        self.N1=0
        self.N2=0
	self.x0=0
        self.x1=0
        self.x2=0
        
        with h5.File(os.path.join(self.chain_path, "restart.h5_0"), mode="r") as f:
	    info=f.require_group('/info')
	    markov=f.require_group('/markov')		
	    #print markov.keys()
	    #print info.keys()
	    print f['info']['scalars'].keys()
	    #print f['markov']['scalars'].keys()	
	
            self.L0 = f['info']['scalars']['L0'][:]
            self.L1 = f['info']['scalars']['L1'][:]
            self.L2 = f['info']['scalars']['L2'][:]
            self.N0 = int(f['info']['scalars']['N0'][:])
            self.N1 = int(f['info']['scalars']['N1'][:])
            self.N2 = int(f['info']['scalars']['N2'][:])
	    self.xmin0 = int(f['info']['scalars']['corner0'][:])
            self.xmin1 = int(f['info']['scalars']['corner1'][:])
            self.xmin2 = int(f['info']['scalars']['corner2'][:])
            self.ncat = int(f['info']['scalars']['NCAT'][:]) 
	

            if(LSS_framework!='BORG'):
                self.kmodes = f['/info/scalars/k_modes'][:]  
                self.nmodes = len(self.kmodes)
        
        #get brefs
            bref=[]	
	    for i in range(self.ncat):
		bref.append(f['info']['scalars']['galaxy_bias_ref_'+str(i)][:])
            self.bias_ref=np.array(bref)
		

    def check_biasref(self):
        return self.bias_ref

    def get_ncat(self):
        return self.ncat	

    def get_mask_spliced(self,msknr,ncpu=0):
        if ncpu>0:
            mskkey = "info.scalars.galaxy_sel_window_" + str(msknr)
            a=rebuild_spliced_h5(os.path.join(self.chain_path, "restart.h5"),[mskkey],32)
            return np.array(a[mskkey][:,:,:,0])
        else:
            print 'Error: need number of processes to read files !'    
            
    def get_mask(self,msknr):
    
	with h5.File(os.path.join(self.chain_path, "restart.h5_0"), mode="r") as f:
		mskkey = "galaxy_sel_window_" + str(msknr)
		mask = f['info']['scalars'][mskkey][:]		    
	return np.array(mask[:,:,:,0])

    def get_data(self,datnr):

	with h5.File(os.path.join(self.chain_path, "restart.h5_0"), mode="r") as f:
		datkey = "galaxy_data_" + str(datnr)
		data = f['info']['scalars'][datkey][:]		    
	return np.array(data)

    def get_data_spliced(self,msknr,ncpu=0):
        if ncpu>0:
            mskkey = "info.scalars.galaxy_data_" + str(msknr)
            a=rebuild_spliced_h5(os.path.join(self.chain_path, "restart.h5"),[mskkey],32)
            return np.array(a[mskkey][:])
        else:
            print 'Error: need number of processes to read files !'    

    def scan_datacube(self,data):
	fig = figure()
	ax = fig.add_subplot(111)
	plt.jet()
	tracker = IndexTracker(ax, data)
	fig.canvas.mpl_connect('key_press_event', tracker.onscroll)
	show()
        
                
    def get_2d_marginal(self,attribute_a='s_field',attribute_b='s_field',id_a=None,id_b=None, first_sample=0,last_sample=1000):
        print '-'*60
        print 'Estimate 2d marginals for parameters ', attribute_a, ' and ', attribute_b , ' for ' , self.LSS_framework, ' run!'
        print '-'*60
        
        if(id_a==None or id_b==None):
            print "Error: no index chosen"
            return -1
            
        #2) collect chain
        samples_a = []
        samples_b = []
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, 1):
                d   = a[attribute_a][:]
                e   = a[attribute_b][:]
                
                samples_a.append(d[id_a])
                samples_b.append(e[id_b])
        
        H, xedges, yedges = np.histogram2d(samples_a, samples_b)
        
        return xedges,yedges, H
        
    def get_cross_corcoeff(self,attribute_a='s_field',attribute_b='s_field',id_a=None,id_b=None, first_sample=0,last_sample=1000):
        print '-'*60
        print 'Estimate 2d marginals for parameters ', attribute_a, ' and ', attribute_b , ' for ' , self.LSS_framework, ' run!'
        print '-'*60
        
        if(id_a==None or id_b==None):
            print "Error: no index chosen"
            return -1
            
        #2) collect chain
        samples_a = []
        samples_b = []
        
        nelements_a = len(id_a[0])
        nelements_b = len(id_b[0])
        
        mu_a    = np.zeros(nelements_a)
        var_a   = np.zeros(nelements_a)
        
        mu_b    = np.zeros(nelements_b)
        var_b   = np.zeros(nelements_b)
        nn=1
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, 1):
                d   = a[attribute_a][:]
                e   = a[attribute_b][:]
                
                aux_a   = d[id_a]
                aux_b   = e[id_b]
                
                mu_a  = (nn-1.)/float(nn)*mu_a +1./float(nn)*aux_a                
                if(nn>1): 
                    aux     = (mu_a-aux_a)**2
                    var_a     = (nn-1.)/nn*var_a+1./(nn-1)*aux
                
                mu_b  = (nn-1.)/float(nn)*mu_b +1./float(nn)*aux_b                
                if(nn>1): 
                    aux     = (mu_b-aux_b)**2
                    var_b     = (nn-1.)/nn*var_b+1./(nn-1)*aux
                
                samples_a.append(aux_a)
                samples_b.append(aux_b)
                nn+=1
        
        pc= np.zeros((nelements_a,nelements_b))
        cnt=0
        for n in range(nn-1):
                x=samples_a[n]
                y=samples_b[n]
                pc += np.multiply.outer(x-mu_a, y-mu_b)
                cnt+=1
        
        return pc/float(cnt) #/np.sqrt(var_a*var_b)    
        
    
    def get_trace(self,attribute='s_field',element_id=None, first_sample=0,last_sample=1000):
        print '-'*60
        print 'Record trace for parameters ', attribute , ' for ' , self.LSS_framework, ' run!'
        print '-'*60
    
        '''
        if(element_id==None):
            print "Error: no list of indices provided"
            return -1
        '''
        
        #1) collect chain
        samples = []
          
        nn=1
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, 1):
                d   = a[attribute][:]
                if (element_id!=None):
                    samples.append(d[element_id])
                else:
                    samples.append(d)          
                nn+=1
                
        return samples
    
    
    def get_corrlength(self,attribute='s_field',element_id=None,nlength=100, first_sample=0,last_sample=1000):
        print '-'*60
        print 'Estimate correlation length for parameters ', attribute , ' for ' , self.LSS_framework, ' run!'
        print '-'*60
    
        if(element_id==None):
            print "Error: no list of indices provided"
            return -1
            
        if(nlength>last_sample-first_sample):
            print "Warning: Chain setting not long enough set nlength to last_sample"
            nlength = last_sample-first_sample -1

        nelements = len(element_id[0])
    
        #1) calculate mean and variance
        mu  = np.zeros(nelements)
        var = np.zeros(nelements)
        
        #2) collect chain
        samples = []

        nn=1
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, 1):
                d   = a[attribute][:]

		print np.shape(d)
          
                mu  = (nn-1.)/float(nn)*mu +1./float(nn)*d[element_id]                
                if(nn>1): 
                    aux     = (mu-d[element_id])**2
                    var     = (nn-1.)/nn*var+1./(nn-1)*aux
                
                samples.append(d[element_id])
                
                nn+=1
        

        cl      = np.zeros((nlength,nelements))
        cl_count= np.zeros(nlength)
   
        for i in range(nlength):
            for j in range(len(samples)-i):
	            cl[i]+= (samples[j]-mu)*(samples[j+i]-mu)/var
	            cl_count[i] +=1.;

        for i in range(nlength):
            cl[i]/=cl_count[i]
       
        return np.array(range(nlength)), cl

    def print_job(self,msg):
        print('-'*60)
        print(msg)
        print('-'*60)



    def spectrum_pdf(self, first_sample=0, last_sample=-1, sample_steps=10, gridsize=1000, Pmin=None, Pmax=None):
        
        P = np.zeros((gridsize, Npk), dtype=np.float64)
        
        if Pmin is None or Pmax is None:
            P0m,P0M = np.inf,0
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                P0m = a['/scalars/powerspectrum'].min()
                P0M = a['/scalars/powerspectrum'].max()
                Pb_m,Pb_M = min(P0m, Pb_m),max(P0M,Pb_M)
            if Pmin is None:
                Pmin = Pb_m
            if Pmax is None:
                Pmax = Pb_M
            
        PP = Pmin*np.exp(np.arange(gridsize)*np.log(Pmax/Pmin))
        N=0
        prior=0
        N_ib = 0.5*(self.Nk+prior)[None,:]
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
            Pk_i = a['/scalars/powerspectrum'][:]
            N_i = self.Nk
            
            _special_spectrum_builder(P, PP, tmp_PP, N_ib, N_i, Pk_i)
            
            N += 1

        P /= N
        
        return P

    def get_spherical_slice(self,vdata,nside=32, observer=np.array([0,0,0]),rslice = 150.):

	def RenderSphere(VolumeData3D,image,rslice,observer,Larr,Narr):
    	    print "Rendering Sphere..."

	    NSIDE=hp.npix2nside(len(image))
    
    	    idx=Larr[0]/Narr[0]
    	    idy=Larr[1]/Narr[1]
    	    idz=Larr[2]/Narr[2]
    
    	    for ipix in range(len(image)):
       		#get direction of pixel and calculate unit vectors
		dx,dy,dz=hp.pix2vec(NSIDE, ipix)
	    	d = math.sqrt(dx * dx + dy * dy + dz * dz)
	    	dx = dx / d; dy = dy / d; dz = dz / d # ray unit vector
	    
	    	rayX = observer[0]+rslice*dx; rayY = observer[1]+rslice*dy; rayZ = observer[2]+rslice*dz
	    	rayX /= idx; rayY /= idy; rayZ /= idz 
	    	#find voxel inside box
	    	ix = int(round(rayX))
	    	iy = int(round(rayY))
	    	iz = int(round(rayZ))
	    
	    	image[ipix]=np.nan 
	    	if ix > -1 and ix < Narr[0] \
            		or iy > -1 and iy < Narr[1] \
            		or iz > -1 and iz < Narr[2]:
	    
	            	jx = (ix+1) % Narr[0];
	            	jy = (iy+1) % Narr[1];
	            	jz = (iz+1) % Narr[2];
	            	rx = (rayX - ix);
	            	ry = (rayY - iy);
	            	rz = (rayZ - iz);
	    
	            	qx = 1.-rx;
	            	qy = 1.-ry;
	            	qz = 1.-rz;
	            	val = VolumeData3D[ix,iy,iz] * qx * qy * qz +VolumeData3D[ix,iy,jz] * qx * qy * rz +VolumeData3D[ix,jy,iz] * qx * ry * qz +VolumeData3D[ix,jy,jz] * qx * ry * rz +VolumeData3D[jx,iy,iz] * rx * qy * qz +VolumeData3D[jx,iy,jz] * rx * qy * rz +VolumeData3D[jx,jy,iz] * rx * ry * qz +VolumeData3D[jx,jy,jz] * rx * ry * rz;
	            	image[ipix]=val

		print '\r'+str(100 * ipix / (len(image) - 1)).zfill(3) + "%"

	
        obs = np.array([observer[0]-self.xmin0,observer[1]-self.xmin1,observer[2]-self.xmin2])

	Larr=np.array([self.L0,self.L1,self.L2])
	Narr=np.array([self.N0,self.N1,self.N2]) 
	image = np.zeros(hp.nside2npix(nside))
        RenderSphere(vdata,image,rslice,obs,Larr,Narr)
	return image

    def mean_var_density(self, first_sample=0,last_sample=-1,sample_steps=10):
        self.print_job('Estimate mean and variance of density fields for %s run!' % self.LSS_framework)
    
        
        if(self.LSS_framework=='ARES'):
            mu_i    = np.zeros((self.N0,self.N1,self.N2))
            
            var_i     = np.zeros((self.N0,self.N1,self.N2))
            
            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['s_field'][:]
                mu_i    = (nn-1.)/float(nn)*mu_i +1./float(nn)*d
                
                if(nn>1): 
                    aux = (mu_i-d)**2
                    var_i   = (nn-1.)/nn*var_i+1./(nn-1)*aux
                    
                nn+=1
            
            return mu_i, var_i
            
        elif(self.LSS_framework=='HADES'):
            mu_i    = np.zeros((self.N0,self.N1,self.N2))
            
            var_i     = np.zeros((self.N0,self.N1,self.N2))
            
            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['s_field'][:]
                mu_i    = (nn-1.)/float(nn)*mu_i +1./float(nn)*d
                
                if(nn>1): 
                    aux = (mu_i-d)**2
                    var_i   = (nn-1.)/nn*var_i+1./(nn-1)*aux
                    
                nn+=1
            
            return mu_i, var_i
            
        else:
            mu_i    = np.zeros((self.N0,self.N1,self.N2))
            mu_f    = np.zeros((self.N0,self.N1,self.N2))
            
            var_i     = np.zeros((self.N0,self.N1,self.N2))
            var_f     = np.zeros((self.N0,self.N1,self.N2))
            
            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['s_field'][:]
                mu_i    = (nn-1.)/float(nn)*mu_i +1./float(nn)*d
                
                if(nn>1): 
                    aux = (mu_i-d)**2
                    var_i   = (nn-1.)/nn*var_i+1./(nn-1)*aux
                
                d = a['BORG_final_density'][:]
                mu_f    = (nn-1.)/float(nn)*mu_f +1./float(nn)*d
                
                if(nn>1):
                    aux = (mu_f-d)**2
                    var_f   = (nn-1.)/nn*var_f+1./(nn-1)*aux
                
                nn+=1
                
                
            return mu_i, var_i, mu_f, var_f

    def mean_var_spec(self, first_sample=0,last_sample=-1,sample_steps=10):
        self.print_job('Estimate mean and variance of density fields for %s run!' % self.LSS_framework)
         
        if(self.LSS_framework=='ARES'):
            mu    = np.zeros(self.nmodes)
            var   = np.zeros(self.nmodes)
            
            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['/scalars/powerspectrum'][:]
                mu    = (nn-1.)/float(nn)*mu +1./float(nn)*d
                
                if(nn>1): 
                    aux = (mu-d)**2
                    var   = (nn-1.)/nn*var+1./(nn-1)*aux
                    
                nn+=1
            
            return self.kmodes,mu, var
            
        elif(self.LSS_framework=='HADES'):
            mu    = np.zeros(self.nmodes)
            var   = np.zeros(self.nmodes)
            
            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['/scalars/powerspectrum'][:]
                mu    = (nn-1.)/float(nn)*mu +1./float(nn)*d
                
                if(nn>1): 
                    aux = (mu-d)**2
                    var   = (nn-1.)/nn*var+1./(nn-1)*aux
                    
                nn+=1
            
            return self.kmodes,mu, var
            
        else:
            mu    = np.zeros(self.nmodes)
            var   = np.zeros(self.nmodes)
            
            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['/scalars/powerspectrum'][:]
                mu    = (nn-1.)/float(nn)*mu +1./float(nn)*d
                
                if(nn>1): 
                    aux = (mu-d)**2
                    var   = (nn-1.)/nn*var+1./(nn-1)*aux
                    
                nn+=1
            
            return self.kmodes,mu, var



#+
#   ARES/HADES/BORG Package -- ./scripts/notebooks/src/special_regions.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import numpy as np

def dir_vec(ra,dec):
    #angles in degrre	
    x=np.cos(dec/180.*np.pi)*np.cos(ra/180.*np.pi)
    y=np.cos(dec/180.*np.pi)*np.sin(ra/180.*np.pi)
    z=np.sin(dec/180.*np.pi)
    return np.array([x,y,z])



#set special sites
special_coords = {}
special_coords['home']={'ra': 0.95,'dec': 0.98,'z': 0.0, 'rc' : 400., 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['coma']={'ra': 194.95,'dec': 27.98,'z': 0.0232, 'rc' : 40., 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['shapley']={'ra': 202.011,'dec': -31.493,'z': 0.0480, 'rc' :  100. *0.68 , 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['coronaborealis']={'ra': 232.0000,'dec': 28.8833,'z': 0.065, 'rc' : 10, 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['bootesvoid']={'ra': 215.0000,'dec': 26.000,'z': 0.05, 'rc' : 12, 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['hydra']={'ra': 158.68749975,'dec': -27.23192123,'z': np.nan, 'rc' : 12, 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['cetus']={'ra': 217.40988269,'dec': -69.94391168,'z': np.nan, 'rc' : 12, 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['vela']={'ra': 141.17998324,'dec': -50.57228815,'z': np.nan, 'rc' : 12, 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['norma']={'ra': 243.5936928,'dec': -60.85205904,'z': np.nan, 'rc' : 12, 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['A3158']={'ra': 55.87672518,'dec': -53.4820204,'z': np.nan, 'rc' : 12, 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['hydra-cen']={'ra': 200.02772883,'dec': -53.81840285,'z': np.nan, 'rc' : 12, 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }
special_coords['horologium']={'ra': 6.74460262,'dec': -49.70333177,'z': np.nan, 'rc' : 12, 'mu_mass' : 1, 'var_mass' : 1, 'mu_delta' : 1, 'var_delta' : 1 }


def get_objpos_range(objname,cosmolo={'omega_M_0' : 0.307, 'omega_lambda_0' : 0.693, 'h' : 0.6777}):
    from astropy.cosmology import LambdaCDM
    cosmo = LambdaCDM(H0=100.*cosmolo['h'], Om0=cosmolo['omega_M_0'], Ode0=cosmolo['omega_lambda_0'])
    
    
    ra   = special_coords[objname]['ra']
    dec  = special_coords[objname]['dec']
    z    = special_coords[objname]['z']
    rc   = special_coords[objname]['rc'] # units Mpc/h

    dcom = np.array(cosmo.comoving_distance(z).value)*cosmolo['h']

    d = dir_vec(ra,dec)
    pos_SSC = dcom * d
    return pos_SSC, rc


# I just want a switch....can't live without it
class switch(object):
    def __init__(self, value):
        self.value = value
        self.fall = False

    def __iter__(self):
        """Return the match method once, then stop"""
        yield self.match
        raise StopIteration
    
    def match(self, *args):
        """Indicate whether or not to enter a case suite"""
        if self.fall or not args:
            return True
        elif self.value in args: # changed for v1.5, see below
            self.fall = True
            return True
        else:
            return False



def sky_region(region,ra,dec):

    for case in switch(region):
        if case('R1'):
            #SDSS-NGC Withburn & Shanks 2014
            return np.where( (ra>150.)*(ra<220.)*(dec>0.)*(dec<50.)), 3072.38*(np.pi/180.)**2
            break
        if case('R2'):
            #6dFGS-SGC Withburn & Shanks 2014
            return np.where( ((ra>330.)*(ra<360.)*(dec<0)*(dec>-50.)) + ((ra>0.)*(ra<50.)*(dec<0)*(dec>-50.))), 3511.29*(np.pi/180.)**2
            break
        if case('R3'):
            #6dFGS-NGC Withburn & Shanks 2014
            return np.where( (ra>150.)*(ra<220.)*(dec<0)*(dec>-40.)), 2578.03*(np.pi/180.)**2
            break
        if case(): # default, could also just omit condition or 'if True'
            print ("Case not known!")
            # No need to break here, it'll stop anyway



import numpy as np
import borg

cons = borg.console()

myprint = lambda x: cons.print_std(x) if type(x) == str else cons.print_std(
    repr(x))

sigma_noise = 0.1


class MyLikelihood(borg.likelihood.BaseLikelihood):
    def __init__(self, fwd, N, L):
        myprint(f" Init {N}, {L}")
        super().__init__(fwd, N, L)
        self.comm = fwd.getCommunicator()

    def initializeLikelihood(self, state):
        myprint("Init likelihood")
        self.data = state['galaxy_data_0']
        state.newArray3d("my_density_field", self.data.shape[0],
                         self.data.shape[1], self.data.shape[2], True)

    def updateMetaParameters(self, state):
        cpar = state['cosmology']
        myprint(f"Cosmology is {cpar}")
        self.getForwardModel().setCosmoParams(cpar)

    def generateMockData(self, s_hat, state):

        fwd = self.getForwardModel()
        output = np.zeros(fwd.getOutputBoxModel().N)
        fwd.forwardModel_v2(s_hat)
        fwd.getDensityFinal(output)

        state['galaxy_data_0'][:] = output + np.random.normal(
            size=output.shape) * sigma_noise
        state['my_density_field'][:] = output
        like = ((state['galaxy_data_0'][:] - output)**2).sum() / sigma_noise**2
        myprint(
            f"Initial log_likelihood: {like}, var(s_hat) = {np.var(s_hat)}")

    def logLikelihoodComplex(self, s_hat, gradientIsNext):
        fwd = self.getForwardModel()

        output = np.zeros(fwd.getBoxModel().N)
        fwd.forwardModel_v2(s_hat)
        fwd.getDensityFinal(output)
        L = 0.5 * ((output - self.data)**2).sum() / sigma_noise**2
        myprint(f"var(s_hat): {np.var(s_hat)}, Call to logLike: {L}")
        return self.comm.allreduce(L)

    def gradientLikelihoodComplex(self, x_hat):
        fwd = self.getForwardModel()
        output = np.zeros(fwd.getOutputBoxModel().N)
        fwd.forwardModel_v2(x_hat)
        fwd.getDensityFinal(output)
        mygradient = (output - self.data) / sigma_noise**2
        fwd.adjointModel_v2(mygradient)
        mygrad_hat = np.zeros(s_hat.shape, dtype=np.complex128)
        fwd.getAdjointModel(mygrad_hat)
        return mygrad_hat


model = None


@borg.registerGravityBuilder
def build_gravity_model(state, box):
    global model
    chain = borg.forward.ChainForwardModel(box)
    chain.addModel(borg.forward.models.HermiticEnforcer(box))
    chain.addModel(borg.forward.models.Primordial(box, 1.0))
    chain.addModel(borg.forward.models.EisensteinHu(box))
    model = chain
    return chain


@borg.registerLikelihoodBuilder
def build_likelihood(state, info):
    boxm = model.getBoxModel()
    like = MyLikelihood(model, boxm.N, boxm.L)
    #like.initializeLikelihood(state)
    return like


@borg.registerSamplerBuilder
def build_samplers(state, info):
    return []



#+
#   ARES/HADES/BORG Package -- ./extra/hades/libLSS/tests/data/gen_gradient_data.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import h5py as h5
import numpy as np

N=32

numbers = np.random.normal(size=(N,N,N))
with h5.File("gradient_numbers.h5", mode="w") as f:
  f['/random'] = numbers
  f['/random_fft'] = np.fft.rfftn(numbers)



from tqdm import tqdm
import math
import os
import numpy as np
import matplotlib as mpl
import matplotlib.cm as cm
mpl.use('Agg')
import ares_tools as at
import matplotlib.pyplot as plt
import h5py as h5
import numba as nb

startMC=0
suffix="ref"

ss = at.analysis(".")
opts=dict(Nbins=256,range=(0,ss.kmodes.max()))

names=[]
PP=[]
Fmax=0
while True:
    try:
      os.stat("mcmc_%d.h5" % Fmax)
    except:
      break
    names.append(Fmax)
    Fmax += 1
print(Fmax)

def handle_likelihood():

  def bias_func(rho, bias):
    a=ne.evaluate('exp(b0*log(rho) - b1*rho**(-b2))', dict(b0=bias[0],b1=bias[1],b2=bias[2],rho=rho))
    return a 
  
  
  @nb.jit(parallel=True,nopython=True)
  def compute_likelihood(S,nmean,bias,density,data):
    N0,N1,N2 = density.shape
    alpha,r0,eps = bias
    L = 0
    for p in nb.prange(N0*N1*N2):
          k = p % N2
          j = (p//N2) % N1
          i = (p//N2//N1)
          if S[i,j,k] <= 0:
            continue
  
          rho = 1+density[i,j,k]+1e-6
          x = r0*rho**(-eps)
          lrho = math.log(rho)
          rho_g = nmean * rho**alpha * math.exp(-x)
          log_rho_g = math.log(nmean) + lrho*alpha - x 
          lam = S[i,j,k] * rho_g
          log_lam = math.log(S[i,j,k]) + log_rho_g
          L += data[i,j,k]*log_lam - lam
  
    return L
  
  try:
    Likelihood = list(np.load("Like_%s.npy" % suffix))
    loc_names = names[len(Likelihood):]
  except:
    Likelihood = []
    loc_names = list(names)
  print(loc_names)
  if len(loc_names) == 0:
    return
  
  data = []
  selection = []
  for d_no in range(16):
    print("Load data %d" % (d_no,))
    data.append(ss.get_data(d_no))
    selection.append(ss.get_mask(d_no))
  
  for mc_id in tqdm(loc_names):
    with h5.File("mcmc_%d.h5" % mc_id, mode="r") as f:
      density = f['/scalars/BORG_final_density'][...]
      L=[]
      for i,(D,S) in enumerate(zip(data,selection)):
        nmean = f['/scalars/galaxy_nmean_%d' % i][0]
        bias = f['/scalars/galaxy_bias_%d' % i][...]
        L.append( compute_likelihood(S, nmean, bias, density, D) )
      Likelihood.append(L)
  
  np.save("Like.npy", Likelihood)


def handle_power():
  Pref = ss.rebin_power_spectrum(startMC, **opts)
  try:
    data = np.load("power_%s.npz" % suffix,allow_pickle=True)
    print("Found previous run")
    loc_names = names[len(data['P']):]
    print(loc_names)
    PP = list(data['P'])
  except Exception as exc:
    print(exc)
    print("no previous run")
    PP = []
    loc_names = list(names)
  print(loc_names)
  if len(loc_names) == 0:
    return
  
  for i in tqdm(loc_names):
    PP.append(ss.compute_power_shat_spectrum(i, **opts))
  
  bins = 0.5*(Pref[2][1:]+Pref[2][:-1])
  
  np.savez("power_%s.npz" % suffix, bins=bins, P=PP, startMC=startMC, Fmax=Fmax, Pref=Pref)
  

#handle_likelihood()
handle_power()



#+
#   ARES/HADES/BORG Package -- ./scripts/migrate_restart.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import os
import h5py as h5
import errno
from ares_tools import rebuild_spliced_h5

def detect_ncpus(path):
  ncpu = 0
  try:
    while True:
      with open("%s_%d" % (path,ncpu), mode= "rb") as f:
        ncpu += 1
  except IOError as e:
    if e.errno != errno.ENOENT:
      raise e
      
  return ncpu
  
def detect_job(path):
  
  obj_list=[]
  group_list=[]
  
  def _handle_item(name, obj):
    if isinstance(obj, h5.Group):
      group_list.append(name)
    if not isinstance(obj, h5.Dataset):
      return

    obj_list.append(name)
      
  with h5.File("%s_0" % path, mode="r") as f:  
    f.visititems(_handle_item)

  return obj_list,group_list

def load_nonarray(path, outpath, objlist, group_list):
  arr = {}
  with h5.File(path, mode="r") as f, h5.File(outpath, mode="w") as of:
    for g in group_list:
      if g[:4] == 'info':
        newname = g[4:]
      elif g[:6] == 'markov':
        newname = g[6:]
      if len(newname)==0:
        continue
      if newname in of:
        continue
      print("Create group %s" % newname)
      of.create_group(newname)
    for oname in objlist:
      print("Loading %s..." % oname)
      if oname[:4] == 'info':
        newname = oname[4:]
      elif oname[:6] == 'markov':
        newname = oname[6:]
      else:
        print("Beuh ! " + oname)
        abort

      if oname == '/info/scalars/BORG_version':
        of[newname] = np.array([f[oname][0]], dtype='S')
      else:
        of[newname] = f[oname][:]
  
def migrate(path, newpath):

  ncpu = detect_ncpus(path)
  elem_list,group_list = detect_job(path)
  
  print("Loading spliced arrays")  
  for n in range(ncpu):
    load_nonarray("%s_%d" % (path,n), "%s_%d" % (newpath,n), elem_list, group_list)
  
  
if __name__=="__main__":
   migrate("./restart.h5", "./new/restart.h5")



#+
#   ARES/HADES/BORG Package -- ./scripts/lic_plot/setup.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy 

setup(
    cmdclass = {'build_ext': build_ext},
    ext_modules = [
	Extension("lic_internal", ["lic_internal.pyx"],
			include_dirs=[numpy.get_include()])
	],
)




#+
#   ARES/HADES/BORG Package -- ./scripts/dump_initial_field.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import pylab as plt
import h5py
import numpy as np
from read_all_h5 import explore_chain


chain_path="."

f = h5py.File('restart.h5_0', "r+")
print list(f['/scalars'])

xmin0 = f['/scalars/corner0'][:]
xmin1 = f['/scalars/corner1'][:]
xmin2 = f['/scalars/corner2'][:]

L0 = f['/scalars/L0'][:]
L1 = f['/scalars/L1'][:]
L2 = f['/scalars/L2'][:]

cosmology=f['/scalars/cosmology'][:]

xmin=np.array([xmin0,xmin1,xmin2])
L=np.array([L0,L1,L2])

outdir ='/scratch/jasche/'

for i,a in explore_chain(chain_path, 5000,8900, 10):
  d = a['s_field'][:]
  fname = outdir+'borg_ic_2m++_'+ str(i)
  print "Saving file : ", fname+'.npz'
  np.savez(fname, 
         ICfield = d, 
         BoxLength = L, 
         posmin = xmin, 
         omega_r = cosmology[0][0],
         omega_k = cosmology[0][1],
         omega_m = cosmology[0][2],
         omega_b = cosmology[0][3],
         omega_q = cosmology[0][4],
         w = cosmology[0][5],
         n_s = cosmology[0][6],
         wprime = cosmology[0][7],
         sigma8 = cosmology[0][8],
         h100 = cosmology[0][10],
         beta = cosmology[0][11])
   
data = np.load(fname+'.npz')

print data.keys()

plt.imshow(data['ICfield'][:,:,128])
plt.show()





#+
#   ARES/HADES/BORG Package -- ./scripts/ares_tools/analysis/analysis.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
from ..read_all_h5 import explore_chain, rebuild_spliced_h5
import errno
import h5py as h5
import numpy as np
import healpy as hp
import numexpr as ne
import os
import math
from pylab import *
import matplotlib.pyplot as plt
from matplotlib.pyplot import figure, show

#import numba
'''
@numba.jit
def _special_spectrum_builder(P, PP, tmp, alpha, N_i, Pk_i):

    for i in range(Pk_i.size):
        for j in range(PP.size):
            v = PP[j]**(-alpha)*exp(- N_i[i]*Pk_i[i] / (2 * PP[j]))
            tmp[j] = v
            total += v

        for j in range(PP.size):
            P[j] += tmp[j] / total
 '''


 #ARES/HADES/BORG image scanning class
class IndexTracker:
    def __init__(self, ax, X):
        self.ax = ax
        self.ax.set_title('use up/down keys to navigate images')

        self.X = X
        rows,cols,self.slices = X.shape
        self.ind  = self.slices/2

        cmax=X.max()
        cmin=X.min()

        self.im = self.ax.imshow(self.X[:,self.ind,:],vmax=cmax,vmin=cmin)
        self.update()

    def onscroll(self, event):
        #print ("%s " % (event.key))
        if event.key=='up':
            self.ind = np.clip(self.ind+1, 0, self.slices-1)
        else:
            self.ind = np.clip(self.ind-1, 0, self.slices-1)


        self.update()

    def update(self):
        self.im.set_data(self.X[:,self.ind,:])
        self.ax.set_ylabel('slice %s'%self.ind)
        self.im.axes.figure.canvas.draw()


def detect_ncpus(path):
  ncpu = 0
  try:
    while True:
      with open("%s_%d" % (path,ncpu), mode= "rb") as f:
        ncpu += 1
  except IOError as e:
    if e.errno != errno.ENOENT:
      raise e

  return ncpu



#ARES/HADES/BORG analysis class
class analysis:


    def __init__(self, chain_path='.', LSS_framework='ARES', start=0, step=1):
        self.chain_path         = chain_path
        self.LSS_framework      = LSS_framework

        self.description = "This Class is part of the ARES/HADES/BORG analysis framework"
        self.author = "Copyright (C) 2009-2017 Jens Jasche \n Copyright (C) 2014-2017 Guilhem Lavaux"

        #get chain setup

        self.L0=0
        self.L1=0
        self.L2=0
        self.N0=0
        self.N1=0
        self.N2=0
        self.x0=0
        self.x1=0
        self.x2=0

        self.ncpus = detect_ncpus(os.path.join(self.chain_path, "restart.h5"))

        self.mcmc_list=[]
        Fmax=start
        while True:
            try:
              os.stat("mcmc_%d.h5" % Fmax)
            except:
              break
            self.mcmc_list.append(Fmax)
            Fmax += step
        

        with h5.File(os.path.join(self.chain_path, "restart.h5_0"), mode="r") as f:
            #print markov.keys()
            #print info.keys()
            #print f['scalars'].keys()

            self.L0 = f['scalars']['L0'][0]
            self.L1 = f['scalars']['L1'][0]
            self.L2 = f['scalars']['L2'][0]
            self.N0 = int(f['scalars']['N0'][:])
            self.N1 = int(f['scalars']['N1'][:])
            self.N2 = int(f['scalars']['N2'][:])
            self.xmin0 = int(f['scalars']['corner0'][:])
            self.xmin1 = int(f['scalars']['corner1'][:])
            self.xmin2 = int(f['scalars']['corner2'][:])
            self.ncat = int(f['scalars']['NCAT'][:])

            if(LSS_framework!='BORG'):
                self.kmodes = f['/scalars/k_modes'][:]
                self.nmodes = len(self.kmodes)


        #get brefs
            #self.k_keys = rebuild_spliced_h5(os.path.join(self.chain_path, "restart.h5"), ["scalars.k_keys"], self.ncpus)["scalars.k_keys"]

            bref=[]
            if LSS_framework != 'VIRBIUS' and LSS_framework != 'LYA':
              for i in range(self.ncat):
                  bref.append(f['scalars']['galaxy_bias_ref_'+str(i)][:])
              self.bias_ref=np.array(bref)

    def _internal_power(self, P, Nbins, range, unit=False):
        if not hasattr(self, '_k'):
          ik0 = np.fft.fftfreq(self.N0, d=self.L0/self.N0)*2*np.pi
          ik1 = np.fft.fftfreq(self.N1, d=self.L1/self.N1)*2*np.pi
          ik2 = np.fft.fftfreq(self.N2, d=self.L2/self.N2)*2*np.pi
          k = self._k = np.sqrt(ik0[:,None,None]**2+ik1[None,:,None]**2+ik2[None,None,:(self.N2//2+1)]**2)
          self._Pw, _ = np.histogram(k, bins=Nbins, range=range)
          Pw = self._Pw
        else:
          k = self._k 
          Pw = self._Pw

        P, b = np.histogram(k, weights=P, bins=Nbins, range=range)
        if not unit:
          P /= self.L0*self.L1*self.L2
        cond = Pw > 0
        P[cond] /= Pw[cond]
        return P, Pw, b 

    def rebin_power_spectrum(self, chain_id, Nbins=100, range=(0,1), unit=False):
        with h5.File(os.path.join(self.chain_path, "mcmc_%d.h5" % (chain_id,)), mode="r") as f:
          P = f['/scalars/powerspectrum'][...] * (self.L0 * self.L1 * self.L2)

        return self._internal_power(P[self.k_keys], Nbins, range, unit=unit)


    def compute_power_spectrum_mock(self, Nbins=100, unit=False, range=(0,1)):
        with h5.File(os.path.join(self.chain_path, "mock_data.h5"), mode="r") as f:
          shat = f['/scalars/s_hat_field'][...]
        return self._internal_power(ne.evaluate('real(s)**2+imag(s)**2', dict(s=shat)), Nbins, range, unit=unit)

    def compute_power_spectrum_galaxydata(self, Nbins=100, range=(0,1)):
        with h5.File(os.path.join(self.chain_path, "mock_data.h5"), mode="r") as f:
          # FFT galaxy data  
          dV = self.L0*self.L1*self.L2/(self.N0*self.N1*self.N2)
          fd = np.fft.rfftn(f['/scalars/galaxy_data_0'][...])*dV
          # remove zero mode
          fd[0][0][0] = 0. + 0.j
        return self._internal_power(ne.evaluate('real(s)**2+imag(s)**2', dict(s=fd)), Nbins, range)

    def compute_power_shat_spectrum(self, chain_id, Nbins=100, unit=False, range=(0,1)):

        dV = self.L0*self.L1*self.L2/(self.N0*self.N1*self.N2)

        with h5.File(os.path.join(self.chain_path, "mcmc_%d.h5" % (chain_id,)), mode="r") as f:
          if '/scalars/s_hat_field' in f:
            shat = f['/scalars/s_hat_field'][...]
          else:
            shat = np.fft.rfftn(f['/scalars/s_field'][...])*dV

        return self._internal_power(ne.evaluate('real(s)**2+imag(s)**2', dict(s=shat)), Nbins, range, unit=unit)

    def compute_power_shat_cross_spectrum(self, chain_id1, chain_id2,
                                          Nbins=100, unit=False, range=(0,1)):

        dV = self.L0*self.L1*self.L2/(self.N0*self.N1*self.N2)

        with h5.File(os.path.join(self.chain_path, "mcmc_%d.h5" % (chain_id1,)), mode="r") as f:
          if '/scalars/s_hat_field' in f:
            shat1 = f['/scalars/s_hat_field'][...]
          else:
            shat1 = np.fft.rfftn(f['/scalars/s_field'][...])*dV

        with h5.File(os.path.join(self.chain_path, "mcmc_%d.h5" % (chain_id2,)), mode="r") as f:
          if '/scalars/s_hat_field' in f:
            shat2 = f['/scalars/s_hat_field'][...]
          else:
            shat2 = np.fft.rfftn(f['/scalars/s_field'][...])*dV

        return self._internal_power(ne.evaluate('real(s1)*real(s2)+imag(s1)*imag(s2)', dict(s1=shat1,s2=shat2)), Nbins, range, unit=unit)
    
        #return self._internal_power(ne.evaluate('real(s)**2+imag(s)**2', dict(s=shat)), Nbins, range, unit=unit)

    # maybe 'unit' argument is not sensible here...
    def compute_power_spectrum_finaldensity(self, chain_id, Nbins=100, unit=False, range=(0,1)):

        dV = self.L0*self.L1*self.L2/(self.N0*self.N1*self.N2)

        with h5.File(os.path.join(self.chain_path, "mcmc_%d.h5" % (chain_id,)), mode="r") as f:
            fd = np.fft.rfftn(f['/scalars/BORG_final_density'][...]) * dV

        return self._internal_power(ne.evaluate('real(s)**2+imag(s)**2', dict(s=fd)), Nbins, range, unit=unit)

    # compute power spectrum of real-space field from given path and field name
    def compute_power_shat_spectrum_file(self, path, fieldname='phases', Nbins=100, unit=False, range=(0,1)):

        dV = self.L0*self.L1*self.L2/(self.N0*self.N1*self.N2)

        with h5.File(path, mode="r") as f:
          if fieldname in f:
              # do FFT
              shat = np.fft.rfftn(f[fieldname][...])*dV
              # remove zero mode
              shat[0][0][0] = 0. + 0.j
          else:
              print("No field '%s' found in file." % fieldname)

        return self._internal_power(ne.evaluate('real(s)**2+imag(s)**2', dict(s=shat)), Nbins, range, unit=unit)

    def check_biasref(self):
        return self.bias_ref

    def get_ncat(self):
        return self.ncat

    def get_mask(self,msknr):
        selkey = "scalars.galaxy_sel_window_%s"%msknr
        return \
             rebuild_spliced_h5(
                 os.path.join(
                    self.chain_path,"restart.h5"
                 ),
                 [selkey],
                 self.ncpus
             )[selkey] 

    def get_data(self,datnr):
        datkey = "scalars.galaxy_data_%s"%datnr
        return \
             rebuild_spliced_h5(
                 os.path.join(
                    self.chain_path,"restart.h5"
                 ),
                 [datkey],
                 self.ncpus
             )[datkey]

    def scan_datacube(self,data):
        fig = figure()
        ax = fig.add_subplot(111)
        plt.jet()
        tracker = IndexTracker(ax, data)
        fig.canvas.mpl_connect('key_press_event', tracker.onscroll)
        show()


    def get_2d_marginal(self,attribute_a='s_field',attribute_b='s_field',id_a=None,id_b=None, first_sample=0,last_sample=1000):
        print( '-'*60)
        print( 'Estimate 2d marginals for parameters ', attribute_a, ' and ', attribute_b , ' for ' , self.LSS_framework, ' run!')
        print( '-'*60)

        if(id_a==None or id_b==None):
            print( "Error: no index chosen")
            return -1

        #2) collect chain
        samples_a = []
        samples_b = []
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, 1):
            d   = a[attribute_a][:]
            e   = a[attribute_b][:]

            samples_a.append(d[id_a])
            samples_b.append(e[id_b])

        H, xedges, yedges = np.histogram2d(samples_a, samples_b)

        return xedges,yedges, H

    def get_cross_corcoeff(self,attribute_a='s_field',attribute_b='s_field',id_a=None,id_b=None, first_sample=0,last_sample=1000):
        print( '-'*60)
        print( 'Estimate 2d marginals for parameters ', attribute_a, ' and ', attribute_b , ' for ' , self.LSS_framework, ' run!')
        print( '-'*60)

        if(id_a==None or id_b==None):
            print("Error: no index chosen")
            return -1

        #2) collect chain
        samples_a = []
        samples_b = []

        nelements_a = len(id_a[0])
        nelements_b = len(id_b[0])

        mu_a    = np.zeros(nelements_a)
        var_a   = np.zeros(nelements_a)

        mu_b    = np.zeros(nelements_b)
        var_b   = np.zeros(nelements_b)
        nn=1
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, 1):
            d   = a[attribute_a][:]
            e   = a[attribute_b][:]

            aux_a   = d[id_a]
            aux_b   = e[id_b]

            mu_a  = (nn-1.)/float(nn)*mu_a +1./float(nn)*aux_a
            if(nn>1):
                aux     = (mu_a-aux_a)**2
                var_a     = (nn-1.)/nn*var_a+1./(nn-1)*aux

            mu_b  = (nn-1.)/float(nn)*mu_b +1./float(nn)*aux_b
            if(nn>1):
                aux     = (mu_b-aux_b)**2
                var_b     = (nn-1.)/nn*var_b+1./(nn-1)*aux

            samples_a.append(aux_a)
            samples_b.append(aux_b)
            nn+=1

        pc= np.zeros((nelements_a,nelements_b))
        cnt=0
        for n in range(nn-1):
            x=samples_a[n]
            y=samples_b[n]
            pc += np.multiply.outer(x-mu_a, y-mu_b)
            cnt+=1

        return pc/float(cnt) #/np.sqrt(var_a*var_b)


    def get_trace(self,attribute='s_field',element_id=None, first_sample=0,last_sample=1000):
        print( '-'*60)
        print( 'Record trace for parameters ', attribute , ' for ' , self.LSS_framework, ' run!')
        print( '-'*60)

        '''
        if(element_id==None):
            print "Error: no list of indices provided"
            return -1
        '''

        #1) collect chain
        samples = []

        nn=1
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, 1):
            d   = a[attribute][:]
            if (element_id!=None):
                samples.append(d[element_id])
            else:
                samples.append(d)
            nn+=1

        return samples


    def get_corrlength(self,attribute='s_field',element_id=None,nlength=100, first_sample=0,last_sample=1000):
        print( '-'*60)
        print( 'Estimate correlation length for parameters ', attribute , ' for ' , self.LSS_framework, ' run!')
        print( '-'*60)

        if(element_id==None):
            print( "Error: no list of indices provided")
            return -1

        if(nlength>last_sample-first_sample):
            print("Warning: Chain setting not long enough set nlength to last_sample")
            nlength = last_sample-first_sample -1

        nelements = len(element_id[0])

        #1) calculate mean and variance
        mu  = np.zeros(nelements)
        var = np.zeros(nelements)

        #2) collect chain
        samples = []

        nn=1
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, 1):
            d   = a[attribute][:]

            print( np.shape(d))

            mu  = (nn-1.)/float(nn)*mu +1./float(nn)*d[element_id]
            if(nn>1):
                aux     = (mu-d[element_id])**2
                var     = (nn-1.)/nn*var+1./(nn-1)*aux

            samples.append(d[element_id])

            nn+=1


        cl      = np.zeros((nlength,nelements))
        cl_count= np.zeros(nlength)

        for i in range(nlength):
            for j in range(len(samples)-i):
                cl[i]+= (samples[j]-mu)*(samples[j+i]-mu)/var
                cl_count[i] +=1.;

        for i in range(nlength):
            cl[i]/=cl_count[i]

        return np.array(range(nlength)), cl

    def print_job(self,msg):
        print('-'*60)
        print(msg)
        print('-'*60)


    def build_power_spectrum_chain(Nbins=256):
      opts=dict(Nbins=Nbins,range=(0,self.kmodes.max()))
      #FIXME: Do not use the first element
      Pref = self.rebin_power_spectrum(self.mcmc_list[0], **opts)
      try:
        data = np.load("power_%s.npz" % suffix)
        loc_names = names[len(data['P']):]
        PP = list(data['P'])
      except:
        PP = []
        loc_names = list(names)
      print(loc_names)
      if len(loc_names) == 0:
        return
      
      for i in pb.ProgressBar()(loc_names):
        PP.append(ss.compute_power_shat_spectrum(i, **opts))
      
      bins = 0.5*(Pref[2][1:]+Pref[2][:-1])
      
      np.savez("power_%s.npz" % suffix, bins=bins, P=PP, startMC=startMC, Fmax=Fmax, Pref=Pref)


    def spectrum_pdf(self, first_sample=0, last_sample=-1, sample_steps=10, gridsize=1000, Pmin=None, Pmax=None):

        P = np.zeros((gridsize, Npk), dtype=np.float64)

        if Pmin is None or Pmax is None:
            P0m,P0M = np.inf,0
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                P0m = a['/scalars/powerspectrum'].min()
                P0M = a['/scalars/powerspectrum'].max()
                Pb_m,Pb_M = min(P0m, Pb_m),max(P0M,Pb_M)
            if Pmin is None:
                Pmin = Pb_m
            if Pmax is None:
                Pmax = Pb_M

        PP = Pmin*np.exp(np.arange(gridsize)*np.log(Pmax/Pmin))
        N=0
        prior=0
        N_ib = 0.5*(self.Nk+prior)[None,:]
        for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
            Pk_i = a['/scalars/powerspectrum'][:]
            N_i = self.Nk

            _special_spectrum_builder(P, PP, tmp_PP, N_ib, N_i, Pk_i)

            N += 1

        P /= N

        return P

    def get_spherical_slice(self,vdata,nside=32, observer=np.array([0,0,0]),rslice = 150.):

        def RenderSphere(VolumeData3D,image,rslice,observer,Larr,Narr):
            print( "Rendering Sphere...")

            NSIDE=hp.npix2nside(len(image))

            idx=Larr[0]/Narr[0]
            idy=Larr[1]/Narr[1]
            idz=Larr[2]/Narr[2]

            for ipix in range(len(image)):
                #get direction of pixel and calculate unit vectors
                dx,dy,dz=hp.pix2vec(NSIDE, ipix)
                d = math.sqrt(dx * dx + dy * dy + dz * dz)
                dx = dx / d; dy = dy / d; dz = dz / d # ray unit vector

                rayX = observer[0]+rslice*dx; rayY = observer[1]+rslice*dy; rayZ = observer[2]+rslice*dz
                rayX /= idx; rayY /= idy; rayZ /= idz
                #find voxel inside box
                ix = int(round(rayX))
                iy = int(round(rayY))
                iz = int(round(rayZ))

                image[ipix]=np.nan
                if ix > -1 and ix < Narr[0] \
                        or iy > -1 and iy < Narr[1] \
                        or iz > -1 and iz < Narr[2]:

                    jx = (ix+1) % Narr[0];
                    jy = (iy+1) % Narr[1];
                    jz = (iz+1) % Narr[2];
                    rx = (rayX - ix);
                    ry = (rayY - iy);
                    rz = (rayZ - iz);

                    qx = 1-rx;
                    qy = 1-ry;
                    qz = 1-rz;
                    val = VolumeData3D[ix,iy,iz] * qx * qy * qz +VolumeData3D[ix,iy,jz] * qx * qy * rz +VolumeData3D[ix,jy,iz] * qx * ry * qz +VolumeData3D[ix,jy,jz] * qx * ry * rz +VolumeData3D[jx,iy,iz] * rx * qy * qz +VolumeData3D[jx,iy,jz] * rx * qy * rz +VolumeData3D[jx,jy,iz] * rx * ry * qz +VolumeData3D[jx,jy,jz] * rx * ry * rz;
                    image[ipix]=val

                print( '\r'+str(100 * ipix / (len(image) - 1)).zfill(3) + "%")


        obs = np.array([observer[0]-self.xmin0,observer[1]-self.xmin1,observer[2]-self.xmin2])

        Larr=np.array([self.L0,self.L1,self.L2])
        Narr=np.array([self.N0,self.N1,self.N2])
        image = np.zeros(hp.nside2npix(nside))
        RenderSphere(vdata,image,rslice,obs,Larr,Narr)
        return image

    def mean_var_density(self, first_sample=0,last_sample=-1,sample_steps=10):
        self.print_job('Estimate mean and variance of density fields for %s run!' % self.LSS_framework)


        if(self.LSS_framework=='ARES'):
            mu_i    = np.zeros((self.N0,self.N1,self.N2))

            var_i     = np.zeros((self.N0,self.N1,self.N2))

            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['s_field'][:]
                mu_i    = (nn-1.)/float(nn)*mu_i +1./float(nn)*d

                if(nn>1):
                    aux = (mu_i-d)**2
                    var_i   = (nn-1.)/nn*var_i+1./(nn-1)*aux

                nn+=1

            return mu_i, var_i

        elif(self.LSS_framework=='HADES'):
            mu_i    = np.zeros((self.N0,self.N1,self.N2))

            var_i     = np.zeros((self.N0,self.N1,self.N2))

            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['s_field'][:]
                mu_i    = (nn-1.)/float(nn)*mu_i +1./float(nn)*d

                if(nn>1):
                    aux = (mu_i-d)**2
                    var_i   = (nn-1.)/nn*var_i+1./(nn-1)*aux

                nn+=1

            return mu_i, var_i

        else:
            mu_i    = np.zeros((self.N0,self.N1,self.N2))
            mu_f    = np.zeros((self.N0,self.N1,self.N2))

            var_i     = np.zeros((self.N0,self.N1,self.N2))
            var_f     = np.zeros((self.N0,self.N1,self.N2))

            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['s_field'][:]
                mu_i    = (nn-1.)/float(nn)*mu_i +1./float(nn)*d

                if(nn>1):
                    aux = (mu_i-d)**2
                    var_i   = (nn-1.)/nn*var_i+1./(nn-1)*aux

                d = a['BORG_final_density'][:]
                mu_f    = (nn-1.)/float(nn)*mu_f +1./float(nn)*d

                if(nn>1):
                    aux = (mu_f-d)**2
                    var_f   = (nn-1.)/nn*var_f+1./(nn-1)*aux

                nn+=1


            return mu_i, var_i, mu_f, var_f

    def mean_var_spec(self, first_sample=0,last_sample=-1,sample_steps=10):
        self.print_job('Estimate mean and variance of density fields for %s run!' % self.LSS_framework)

        if(self.LSS_framework=='ARES'):
            mu    = np.zeros(self.nmodes)
            var   = np.zeros(self.nmodes)

            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['/scalars/powerspectrum'][:]
                mu    = (nn-1.)/float(nn)*mu +1./float(nn)*d

                if(nn>1):
                    aux = (mu-d)**2
                    var   = (nn-1.)/nn*var+1./(nn-1)*aux

                nn+=1

            return self.kmodes,mu, var

        elif(self.LSS_framework=='HADES'):
            mu    = np.zeros(self.nmodes)
            var   = np.zeros(self.nmodes)

            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['/scalars/powerspectrum'][:]
                mu    = (nn-1.)/float(nn)*mu +1./float(nn)*d

                if(nn>1):
                    aux = (mu-d)**2
                    var   = (nn-1.)/nn*var+1./(nn-1)*aux

                nn+=1

            return self.kmodes,mu, var

        else:
            mu    = np.zeros(self.nmodes)
            var   = np.zeros(self.nmodes)

            nn=1
            for i,a in explore_chain(self.chain_path, first_sample,last_sample, sample_steps):
                d = a['/scalars/powerspectrum'][:]
                mu    = (nn-1.)/float(nn)*mu +1./float(nn)*d

                if(nn>1):
                    aux = (mu-d)**2
                    var   = (nn-1.)/nn*var+1./(nn-1)*aux

                nn+=1

            return self.kmodes,mu, var



#+
#   ARES/HADES/BORG Package -- ./scripts/check_gradients.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import matplotlib
matplotlib.use('Agg')
from ares_tools import read_all_h5
import pylab as plt
from sys import argv

file='dump.h5'

if (len(argv) > 1):
    file = argv[1]

print('Reading from %s.' % file)
g=read_all_h5(file)

#ss=16
#step=10

ss=4#8*8
step=5

prior = g.scalars.gradient_array_prior[::ss,:,:].flatten()
prior_ref = g.scalars.gradient_array_prior[::ss,:,:].flatten()

dpr_adj_re= prior.real
dpr_ref_re= prior_ref.real
dpr_adj_im= prior.imag
dpr_ref_im= prior_ref.imag

lh = g.scalars.gradient_array_lh[::ss,:,:].flatten()
lh_ref = g.scalars.gradient_array_lh_ref[::ss,:,:].flatten()

dlh_adj_re= lh.real
dlh_ref_re= 1*lh_ref.real
dlh_adj_im= lh.imag
dlh_ref_im= 1*lh_ref.imag

fig = plt.figure(figsize=(12, 6))
fig.subplots_adjust(left=0.08, right=0.98, bottom=0.1, top=0.95, wspace=0.25, hspace=0.16)

ax1=plt.subplot(2,2,1)             # left subplot in top row
plt.axhline(0.0, color='black', linestyle=':')
plt.plot(dpr_adj_re[::step],'ro',markersize=5.)
plt.plot(dpr_ref_re[::step],color='blue')
ax1.yaxis.get_major_formatter().set_powerlimits((-2, 2))
ax1.xaxis.set_ticklabels('')
plt.ylabel('dPSI_prior_real')

ax2=plt.subplot(2,2,2)             # right subplot in top row
plt.axhline(0.0, color='black', linestyle=':')
gg,=plt.plot(dpr_adj_im[::step],'ro',markersize=5.)
rr,=plt.plot(dpr_ref_im[::step],color='blue')
ax2.legend((gg,rr),('gradient','finite diff'))
ax2.yaxis.get_major_formatter().set_powerlimits((-2, 2))
ax2.xaxis.set_ticklabels('')
plt.ylabel('dPSI_prior_imag')

ax3=plt.subplot(2,2,3)             # left subplot in bottom row
plt.axhline(0.0, color='black', linestyle=':')
plt.plot(dlh_adj_re[::step],'ro',markersize=5.)
plt.plot(dlh_ref_re[::step],color='blue')
ax3.yaxis.get_major_formatter().set_powerlimits((-2, 2))
plt.xlabel('voxel ID')
plt.ylabel('dPSI_likelihood_real')

ax4=plt.subplot(2,2,4)             # right subplot in bottom row
plt.axhline(0.0, color='black', linestyle=':')
plt.plot(dlh_adj_im[::step],'ro',markersize=5.)
plt.plot(dlh_ref_im[::step],color='blue')
ax4.yaxis.get_major_formatter().set_powerlimits((-2, 2))
plt.xlabel('voxel ID')
plt.ylabel('dPSI_likelihood_imag')

plt.savefig('check_gradient.png')

#plt.scatter(dpr_adj_re,dpr_ref_re)
#plt.show()



#+
#   ARES/HADES/BORG Package -- ./libLSS/tests/data/gen_reference_data.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import h5py as h5
import numpy as np

with h5.File("reference_data.h5", mode="w") as f:
  for N in [32]:
    numbers = np.random.normal(size=(N,N,N))
    f[f'/f_size_{N}'] = numbers
    f[f'/c_size_{N}'] = np.fft.rfftn(numbers)



#+
#   ARES/HADES/BORG Package -- ./scripts/lic_plot/lic.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import numpy as np


def lic_flow(vectors,len_pix=10):
    vectors = np.asarray(vectors)
    m,n,two = vectors.shape
    if two!=2:
        raise ValueError

    result = np.zeros((2*len_pix+1,m,n,2),dtype=np.int32) # FIXME: int16?
    center = len_pix
    result[center,:,:,0] = np.arange(m)[:,np.newaxis]
    result[center,:,:,1] = np.arange(n)[np.newaxis,:]

    for i in range(m):
        for j in range(n):
            y = i
            x = j
            fx = 0.5
            fy = 0.5
            for k in range(len_pix):
                vx, vy = vectors[y,x]
                print x, y, vx, vy
                if vx>=0:
                    tx = (1-fx)/vx
                else:
                    tx = -fx/vx
                if vy>=0:
                    ty = (1-fy)/vy
                else:
                    ty = -fy/vy
                if tx<ty:
                    print "x step"
                    if vx>0:
                        x+=1
                        fy+=vy*tx
                        fx=0.
                    else:
                        x-=1
                        fy+=vy*tx
                        fx=1.
                else:
                    print "y step"
                    if vy>0:
                        y+=1
                        fx+=vx*ty
                        fy=0.
                    else:
                        y-=1
                        fx+=vx*ty
                        fy=1.
                if x<0: x=0
                if y<0: y=0
                if x>=n: x=n-1
                if y>=m: y=m-1
                result[center+k+1,i,j,:] = y, x



    return result



#+
#   ARES/HADES/BORG Package -- ./scripts/misc/plot_power.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import read_all_h5
from pylab import *

P=[]
n=[]
b=[]
n1=[]
b1=[]
i=0
while True:
  a = \
   read_all_h5.read_all_h5("mcmc_%d.h5" % i)
  try:
    P.append(a.scalars.powerspectrum)
    n.append(a.scalars.galaxy_nmean_0[0])
    b.append(a.scalars.galaxy_bias_0[0])
    n1.append(a.scalars.galaxy_nmean_1[0])
    b1.append(a.scalars.galaxy_bias_1[0])
  except AttributeError:
    break
  i += 1

k = read_all_h5.read_all_h5("info.h5").scalars.k_modes  
P = np.array(P)

f=figure(1)
clf()
loglog(k[:,None].repeat(P.shape[0],axis=1),P.transpose())

f=figure(2)
clf()
plot(n)




import itertools
from tqdm import tqdm
import borg
import numpy as np

cons = borg.console()

myprint = lambda x: cons.print_std(x) if type(x) == str else cons.print_std(
    repr(x))


def build_gravity_model(box, cpar):
    lpt = borg.forward.models.BorgLpt(box, box, ai=1.0)
    chain = borg.buildDefaultChain(box, cpar, 1.0, lpt)
    vfield = borg.forward.velocity.CICModel(box, lpt)
    return chain, lpt, vfield


if not borg.EMBEDDED and __name__ == "__main__":
    from tqdm import tqdm

    Ng = 8

    # Create the auxiliary objects
    cpar = borg.cosmo.CosmologicalParameters()
    box = borg.forward.BoxModel(100., Ng)

    fwd, lpt, vfield = build_gravity_model(box, cpar)

    s_hat = np.fft.rfftn(np.random.randn(Ng, Ng, Ng) / Ng**(1.5))

    def vfield_like(s_hat):
        fwd.forwardModel_v2(s_hat)
        rho = np.zeros((Ng, Ng, Ng))
        fwd.getDensityFinal(rho)
        vgrid = vfield.getVelocityField()
        return (vgrid**2).sum()

    def vfield_ag(s_hat):
        fwd.forwardModel_v2(s_hat)
        rho = np.zeros((Ng, Ng, Ng))
        fwd.getDensityFinal(rho)
        # The derivative of square is 2 * vector
        vgrid = 2 * vfield.getVelocityField()

        fwd.clearAdjointGradient()
        vfield.computeAdjointModel(vgrid)
        myprint("Calling adjoint")
        # We have to trigger the adjoint computation in any case
        fwd.adjointModel_v2(None)
        myprint("Done with adjoint")
        analytic_gradient = np.zeros((Ng, Ng, Ng // 2 + 1),
                                     dtype=np.complex128)
        fwd.getAdjointModel(analytic_gradient)
        return analytic_gradient

    myprint("Running adjoint")

    num_gradient = np.zeros((Ng, Ng, Ng // 2 + 1), dtype=np.complex128)
    s_hat_epsilon = s_hat.copy()
    cons.setVerboseLevel(5)
    analytic_gradient = vfield_ag(s_hat)
    cons.setVerboseLevel(1)

    epsilon = 0.001
    for i, j, k in tqdm(itertools.product(*map(range, [Ng, Ng, Ng // 2 + 1])),
                        total=Ng * Ng * (Ng // 2 + 1)):
        s_hat_epsilon[i, j, k] = s_hat[i, j, k] + epsilon
        L = vfield_like(s_hat_epsilon)
        s_hat_epsilon[i, j, k] = s_hat[i, j, k] - epsilon
        L -= vfield_like(s_hat_epsilon)
        QQ = L / (2.0 * epsilon)

        s_hat_epsilon[i, j, k] = s_hat[i, j, k] + 1j * epsilon
        L = vfield_like(s_hat_epsilon)
        s_hat_epsilon[i, j, k] = s_hat[i, j, k] - 1j * epsilon
        L -= vfield_like(s_hat_epsilon)
        QQ = QQ + L * 1j / (2.0 * epsilon)

        s_hat_epsilon[i, j, k] = s_hat[i, j, k]

        num_gradient[i, j, k] = QQ

    np.savez("gradients.npz", num=num_gradient, ana=analytic_gradient)



#+
#   ARES/HADES/BORG Package -- ./scripts/quasar/gen_selection_qso.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import cosmolopy as cpy
import numpy as np
import pyfits as pf

cosmo={'omega_M_0':0.30,'omega_lambda_0':0.70,'omega_k_0':0,'h':0.65,'sigma8':0.80}

f = pf.open("DR12Q.fits")
c = f[1].data

Z = c['Z_PIPE']

d = cpy.distance.comoving_distance(Z, **cosmo) * cosmo['h']

Dmax = 8000
Nb = 100
delta = 0.5*Dmax/Nb

d = d[d>100]

H,b = np.histogram(d, range=(0-delta,Dmax-delta),bins=Nb)

b0 = 0.5*(b[1:] + b[0:b.size-1])
H = H.astype(np.float64) / (b[1:]**3 - b[0:b.size-1]**3)

b0max = Dmax


H /= H.max()

with open("quasar_selection.txt", mode="wt") as f:

  f.write("%d %lg\n" % (H.size, b0max))
  for r in H:
    f.write("%lg\n" % r)




from matplotlib import pyplot as plt
import io
import numpy as np
import sys
import mpi4py
import contextlib
import mpi4py.MPI
import _borg

cons=_borg.console()

M = 128
pfact=1.02

class RewriteIO(io.IOBase):
    def __init__(self, original):
        self.original = original
        self.newline = False
        self.comm = mpi4py.MPI.COMM_WORLD
        self.prefix = f'[{self.comm.rank} / {self.comm.size}] '

    def write(self, b):
        if b[0] == '\n':
            return self.original.write(b)
        return self.original.write(self.prefix + b)

    def writelines(self,lines):
        return self.original.writelines(lines)


def make_P(box, delta):
  ik = np.fft.fftfreq(box.N[0], d=box.L[0]/box.N[0] ) *2*np.pi
  kx,ky,kz = np.meshgrid(ik,ik,ik[:(box.N[0]//2 + 1)])
  kn = np.sqrt(kx**2 + ky**2 + kz**2)

  delta_hat = np.fft.rfftn(delta) * box.volume / box.Ntot
  P, bins = np.histogram(kn, weights = np.abs(delta_hat)**2, range=(0,1),bins=50)
  Pw, _ = np.histogram(kn, range=(0,1),bins=50)
  P = P / Pw / box.volume
  return bins, P
  

with contextlib.redirect_stdout(RewriteIO(sys.stdout)):
  
  cons.setVerboseLevel(5)
  
  cosmo_par = _borg.cosmo.CosmologicalParameters()
  cosmo_par.default()
  print(repr(cosmo_par))
  
  cosmo = _borg.cosmo.Cosmology(cosmo_par)
  cpower = _borg.cosmo.CosmoPower(cosmo_par)
  
  box = _borg.forward.BoxModel()
  box.L = (600,600,600)
  box.N = (M,M,M)
  print(box)
  
  # Generate random numbers variance 1 in fourier space
  indelta = np.random.randn(*box.N)
  indelta /= np.sqrt(indelta.size)
  delta_hat = np.fft.rfftn(indelta)

  comm = mpi4py.MPI.COMM_WORLD
  delta_hat = delta_hat[comm.rank*box.N[0]//comm.size : (comm.rank+1)*box.N[0]//comm.size, :, :]
  
  # Generate k grid
  ik = np.fft.fftfreq(box.N[0], d=box.L[0]/box.N[0] ) *2*np.pi
  kx,ky,kz = np.meshgrid(ik,ik,ik[:(box.N[0]//2 + 1)])
  kn = np.sqrt(kx**2 + ky**2 + kz**2)
  kn = kn[comm.rank*box.N[0]//comm.size : (comm.rank+1)*box.N[0]//comm.size, :, :]
  
  # Scale and multiply by power spectrum
  scale = cosmo.d_plus(0.1)/cosmo.d_plus(1.0)
  delta_hat *= scale* np.sqrt(box.L[0]**3 * cpower.power(kn))

  outN = (comm.rank+1)*box.N[0]//comm.size - (comm.rank)*box.N[0]//comm.size, box.N[1], box.N[2]
  
  #
  # Apply LPT
  final_delta_lpt = np.zeros(outN)
  model = _borg.forward.models.BorgLpt(box,ai=0.1, particle_factor=pfact)
  model.holdParticles()
  model.setCosmoParams(cosmo_par)
  model.forwardModel(delta_hat / box.L[0]**3, final_delta_lpt)
  del model

  final_delta_lpt_ngp_quad = np.zeros(outN)
  model = _borg.forward.models.BorgLptNGP_Quad(box,ai=0.1, particle_factor=pfact)
  model.holdParticles()
  model.setCosmoParams(cosmo_par)
  model.forwardModel(delta_hat / box.L[0]**3, final_delta_lpt_ngp_quad)
  del model

  final_delta_lpt_smooth_ngp_quad = np.zeros(outN)
  model = _borg.forward.models.BorgLptSmoothNGP_Quad(box,ai=0.1, particle_factor=pfact)
  model.holdParticles()
  model.setCosmoParams(cosmo_par)
  model.forwardModel(delta_hat / box.L[0]**3, final_delta_lpt_smooth_ngp_quad)
  del model


  P_cic = make_P(box, final_delta_lpt)
  P_ngp_quad = make_P(box, final_delta_lpt_ngp_quad)
  P_smooth_ngp_quad = make_P(box, final_delta_lpt_smooth_ngp_quad)

  kk = (P_cic[0][1:]+P_cic[0][:-1])/2
  fig=plt.figure()
  ax = fig.add_subplot(111)
  ax.loglog(kk,P_ngp_quad[1],label='MNGP QUAD')
  ax.loglog(kk,P_smooth_ngp_quad[1],label='Smooth QUAD')
  ax.loglog(kk,P_cic[1],label='CIC')
  ax.loglog(kk,cpower.power(kk),label='linear')
  ax.legend()
  ax.set_xlabel('$k (h/Mpc)$')
  ax.set_ylabel("$P(k)$")
  ax.set_ylim(1,1e5)

  fig.savefig(f"P_{box.N[0]}.pdf")




#
# Example to run a forward model with PyBORG
#
import borg
import numpy as np


# Setup resolution of the initial mesh
Ng=64
# Box size in Mpc/h
L=100.

# setup the box
bb = borg.forward.BoxModel()
bb.L = L,L,L
bb.N = Ng,Ng,Ng

# Initialize some default cosmology
cosmo = borg.cosmo.CosmologicalParameters()

# Fiducial scale factor to express initial conditions
a0=0.1

chain = borg.forward.ChainForwardModel(bb)
# Add primordial fluctuations
chain.addModel(borg.forward.models.Primordial(bb, a0))
# Add E&Hu transfer function
chain.addModel(borg.forward.models.EisensteinHu(bb))
# Run an LPT model from a=0.0 to af. The ai=a0 is the scale factor at which the IC are expressed
lpt = borg.forward.models.BorgLpt(bb, bb, ai=a0, af=1.0)
chain.addModel(lpt)

# Set the cosmology
chain.setCosmoParams(cosmo)


# Generate white noise: it has to be scaled by 1/N**(3./2) to be one in Fourier
ic = np.random.randn(Ng, Ng, Ng)/np.sqrt(Ng**3)
delta_m = np.zeros((Ng,Ng,Ng))

# RUN!
chain.forwardModel_v2(ic)
chain.getDensityFinal(delta_m)

# Obtain the particles
pp = np.zeros((lpt.getNumberOfParticles(),3))
lpt.getParticlePositions(pp)

print(pp)



#+
#   ARES/HADES/BORG Package -- ./scripts/misc/plot_void.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
from pylab import *
from read_all_h5 import explore_chain

def box2sphere(x,y,z):
 #calculate radii
 r=np.sqrt(x**2+y**2+z**2)
 
 print np.shape(r),np.shape(x)
 
 dec=np.zeros(np.shape(r))
 '''
 foo= np.where(r>0)
 dec[foo]=np.arcsin(z[foo]/r[foo])
 '''
 ra=np.arctan2(y,x)
 
 print np.shape(r),np.shape(ra)
 
 return ra,dec,r


chain_path="."

N = 256 
L = 677.7
Nb = 128 
f = np.sqrt(3)*0.5

ix = np.arange(N)*L/N - 0.5*L

ra,dec,r=box2sphere(ix[:,None,None],ix[None,:,None],ix[None,None,:])

r = np.sqrt(ix[:,None,None]**2 + ix[None,:,None]**2 + ix[None,None,:]**2)

print np.shape(r)

H, b = np.histogram(r, range=(0,f*L), bins=Nb)

Hw_mean=np.zeros(np.shape(H))

cnt=0

mu  = np.zeros(np.shape(H))
var = np.zeros(np.shape(H))

nn=1
for i,a in explore_chain(chain_path, 400,4100, 10):
  d = a['BORG_final_density'][:]

  Hw, b = np.histogram(r, weights=d, range=(0,f*L), bins=Nb)
  Hw /= H
  
  mu    = (nn-1.)/float(nn)*mu +1./float(nn)*Hw
  if(nn>1): 
          aux = (mu-Hw)**2
          var = (nn-1.)/nn*var+1./(nn-1)*aux
  
  nn+=1
  
plot(b[1:], mu, label='average', color='red')

fill_between(b[1:], mu, mu+np.sqrt(var), interpolate=True, color='gray', alpha='0.5')
fill_between(b[1:], mu-np.sqrt(var), mu, interpolate=True, color='gray', alpha='0.5')
fill_between(b[1:], mu, mu+2*np.sqrt(var), interpolate=True, color='darkgray', alpha='0.5')
fill_between(b[1:], mu-2*np.sqrt(var), mu, interpolate=True, color='darkgray', alpha='0.5')


plt.xlabel(r'$r \left[\mathrm{Mpc/h} \right]$')
plt.ylabel(r'$\langle \delta \rangle$')

    
axhline(0.0,lw=1.5, color='black')

#legend()
ylim(-1,1)
gcf().savefig("void.png")



import numpy as np
import _borg

cosmo_par = _borg.cosmo.CosmologicalParameters()
cosmo_par.default()
print(repr(cosmo_par))

cosmo = _borg.cosmo.Cosmology(cosmo_par)
cpower = _borg.cosmo.CosmoPower(cosmo_par)

box = _borg.forward.BoxModel()
box.L = (100,100,100)
box.N = (128,128,128) 
print(box)

# Generate random numbers variance 1 in fourier space
indelta = np.random.randn(*box.N)
indelta /= np.sqrt(indelta.size)
delta_hat = np.fft.rfftn(indelta)

# Generate k grid
ik = np.fft.fftfreq(box.N[0], d=box.L[0]/box.N[0] ) *2*np.pi
kx,ky,kz = np.meshgrid(ik,ik,ik[:(box.N[0]//2 + 1)])
kn = np.sqrt(kx**2 + ky**2 + kz**2)

# Scale and multiply by power spectrum
scale = cosmo.d_plus(0.1)/cosmo.d_plus(1.0)
delta_hat *= scale* np.sqrt(box.L[0]**3 * cpower.power(kn))

#
# Apply LPT
final_delta_lpt = np.zeros(box.N)
model = _borg.forward.models.BorgLpt(box,ai=0.1, particle_factor=2)
model.setCosmoParams(cosmo_par)
model.forwardModel(delta_hat / box.L[0]**3, final_delta_lpt)


biased_density = np.zeros_like(final_delta_lpt)
biased_density_2 = np.zeros_like(final_delta_lpt)
_borg.bias.PowerLawBias().compute(model, 1.0, [1.0,2.0], final_delta_lpt, biased_density)
_borg.bias.BrokenPowerLawBias().compute(model, 1.0, [1.0,2.0,0.1,0.1], final_delta_lpt, biased_density_2)



#+
#   ARES/HADES/BORG Package -- ./scripts/mod_restart_file.py
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+
import h5py
import numpy as np


#set chain
chain_path="."

#set reference file
refdir ='/scratch/jasche/panphasia_run_lc/'
fref=refdir+'restart.h5_0'

#set target file
tardir ='/scratch/jasche/panphasia_run_h/'
ftar=tardir+'restart.h5_0'

#read density files from reference file
print fref
dref = h5py.File(fref, "r")

dref_final_density=dref['/scalars/BORG_final_density']
dref_s_field=dref['/scalars/s_field']
dref_s_hat_field=dref['/scalars/s_hat_field']


dtar = h5py.File(ftar, "r+")
dtar_final_density=dtar['/scalars/BORG_final_density']
dtar_s_field=dtar['/scalars/s_field']
dtar_s_hat_field=dtar['/scalars/s_hat_field']

'''
WARNING: At this point you will irretrievably
modify your restart file!!!!!!
'''


dtar_final_density[...] = dref_final_density[...]
dtar_s_field[...]       = dref_s_field[...]
dtar_s_hat_field[...]   = dref_s_hat_field[...]


dref.close()
dtar.close()



#+
#   ARES/HADES/BORG Package -- ./scripts/lic_plot/lic_internal.pyx
#   Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#   Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
#   
#+

import numpy as np
cimport numpy as np

cdef void _advance(float vx, float vy,
        int* x, int* y, float*fx, float*fy, int w, int h):
    cdef float tx, ty
    if vx>=0:
        tx = (1-fx[0])/vx
    else:
        tx = -fx[0]/vx
    if vy>=0:
        ty = (1-fy[0])/vy
    else:
        ty = -fy[0]/vy
    if tx<ty:
        if vx>=0:
            x[0]+=1
            fx[0]=0
        else:
            x[0]-=1
            fx[0]=1
        fy[0]+=tx*vy
    else:
        if vy>=0:
            y[0]+=1
            fy[0]=0
        else:
            y[0]-=1
            fy[0]=1
        fx[0]+=ty*vx
    if x[0]>=w:
        x[0]=w-1 # FIXME: other boundary conditions?
    if x[0]<0:
        x[0]=0 # FIXME: other boundary conditions?
    if y[0]<0:
        y[0]=0 # FIXME: other boundary conditions?
    if y[0]>=h:
        y[0]=h-1 # FIXME: other boundary conditions?


#np.ndarray[float, ndim=2] 
def line_integral_convolution(
        np.ndarray[float, ndim=3] vectors,
        np.ndarray[float, ndim=2] texture,
        np.ndarray[float, ndim=1] kernel):
    cdef int i,j,k,x,y
    cdef int h,w,kernellen
    cdef int t
    cdef float fx, fy, tx, ty
    cdef np.ndarray[float, ndim=2] result

    h = vectors.shape[0]
    w = vectors.shape[1]
    t = vectors.shape[2]
    kernellen = kernel.shape[0]
    if t!=2:
        raise ValueError("Vectors must have two components (not %d)" % t)
    result = np.zeros((h,w),dtype=np.float32)

    for i in range(h):
        for j in range(w):
            x = j
            y = i
            fx = 0.5
            fy = 0.5
            
            k = kernellen//2
            #print i, j, k, x, y
            result[i,j] += kernel[k]*texture[x,y]
            while k<kernellen-1:
                _advance(vectors[y,x,0],vectors[y,x,1],
                        &x, &y, &fx, &fy, w, h)
                k+=1
                #print i, j, k, x, y
                result[i,j] += kernel[k]*texture[x,y]

            x = j
            y = i
            fx = 0.5
            fy = 0.5
            
            while k>0:
                _advance(-vectors[y,x,0],-vectors[y,x,1],
                        &x, &y, &fx, &fy, w, h)
                k-=1
                #print i, j, k, x, y
                result[i,j] += kernel[k]*texture[x,y]
                    
    return result




# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
# sys.path.insert(0, os.path.abspath('.'))
sys.path.append(os.path.abspath('../sphinx_ext/'))
import datetime
now = datetime.datetime.now()
year = '{:02d}'.format(now.year)

# -- Project information -----------------------------------------------------
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.intersphinx', 'sphinx.ext.autosummary',
    'sphinx.ext.napoleon', 'sphinx_rtd_theme', 'sphinx.ext.mathjax',
    'sphinx.ext.todo', 'nbsphinx',
    'IPython.sphinxext.ipython_console_highlighting', 'sphinx_copybutton',
    'toctree_filter'
]
master_doc = 'index'
source_suffix = '.rst'
rst_prolog = '''
.. |a| replace:: *ARES*
'''

# General information about the project.
project = u'ARES-HADES-BORG'
author = u'the Aquila Consortium'
copyright = u"""
2009-""" + year + """, the Aquila Consortium
"""
#version = "latest"

autosummary_generate = True

todo_include_todos = True

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
#extensions = [
#        'breathe',
#        'exhale',
#]

nbsphinx_execute = 'never'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.inc.rst']
# Excluding the extension .inc.rst avoids compiling "included" rst file
# (otherwise the corresponding .html is produced) and avoids the "duplicate label"
# warning in case a label is found there (Florent Leclercq, 24-10-2020)

#html_extra_path = [os.path.abspath('../_build/html')]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

html_context = {
    'theme_vcs_pageview_mode': 'view&spa=0'
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = [
    'css/custom.css',
]

# --- Breathe/Exhale options

breathe_projects = {"ARES libLSS": "./doxyoutput/xml"}

breathe_default_project = "ARES libLSS"

exhale_args = {
    "containmentFolder": "./api",
    "rootFileName": "library_root.rst",
    "rootFileTitle": "Library API",
    "doxygenStripFromPath": "..",
    "createTreeView": True,
    "exhaleExecutesDoxygen": True,
    "exhaleUseDoxyfile": True
}

primary_domain = 'py'
highlight_language = 'py'

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
    toc_filter_exclude = []
    meta={"bitbucket_url": 'https://www.bitbucket.org/bayesian_lss_team/ares'}
    bitbucket_url='https://www.bitbucket.org/bayesian_lss_team/ares'
else:
    toc_filter_exclude = ["aquila"]

import subprocess

os.environ["ARES_BASE"] = os.path.abspath(os.path.join(os.getcwd(), ".."))
#subprocess.call('doxygen Doxyfile', shell=True)



import io
import numpy as np
import sys
import mpi4py
import contextlib
import borg

cons=borg.console()

class RewriteIO(io.IOBase):
    def __init__(self, original):
        self.original = original
        self.newline = False
        self.comm = mpi4py.MPI.COMM_WORLD
        self.prefix = f'[{self.comm.rank} / {self.comm.size}] '

    def write(self, b):
        if b[0] == '\n':
            return self.original.write(b)
        return self.original.write(self.prefix + b)

    def writelines(self,lines):
        return self.original.writelines(lines)


with contextlib.redirect_stdout(RewriteIO(sys.stdout)):
  
  cons.setVerboseLevel(5)
  
  cosmo_par = borg.cosmo.CosmologicalParameters()
  cosmo_par.default()
  print(repr(cosmo_par))
  
  cosmo = borg.cosmo.Cosmology(cosmo_par)
  cpower = borg.cosmo.CosmoPower(cosmo_par)
  
  box = borg.forward.BoxModel()
  box.L = (100,100,100)
  box.N = (128,128,128)
  print(box)
  
  # Generate random numbers variance 1 in fourier space
  indelta = np.random.randn(*box.N)
  indelta /= np.sqrt(indelta.size)
  delta_hat = np.fft.rfftn(indelta)

  comm = mpi4py.MPI.COMM_WORLD
  delta_hat = delta_hat[comm.rank*box.N[0]//comm.size : (comm.rank+1)*box.N[0]//comm.size, :, :]
  
  # Generate k grid
  ik = np.fft.fftfreq(box.N[0], d=box.L[0]/box.N[0] ) *2*np.pi
  kx,ky,kz = np.meshgrid(ik,ik,ik[:(box.N[0]//2 + 1)])
  kn = np.sqrt(kx**2 + ky**2 + kz**2)
  kn = kn[comm.rank*box.N[0]//comm.size : (comm.rank+1)*box.N[0]//comm.size, :, :]
  
  # Scale and multiply by power spectrum
  scale = cosmo.d_plus(0.1)/cosmo.d_plus(1.0)
  delta_hat *= scale* np.sqrt(box.L[0]**3 * cpower.power(kn))

  outN = (comm.rank+1)*box.N[0]//comm.size - (comm.rank)*box.N[0]//comm.size, box.N[1], box.N[2]
  
  #
  # Apply 2LPT
  final_delta_2lpt = np.zeros(outN)
  model = borg.forward.models.Borg2Lpt(box,ai=0.1, particle_factor=2)
  model.setCosmoParams(cosmo_par)
  model.forwardModel(delta_hat / box.L[0]**3, final_delta_2lpt)
  del model
  
  #
  # Apply LPT
  final_delta_lpt = np.zeros(outN)
  model = borg.forward.models.BorgLpt(box,ai=0.1, particle_factor=2)
  model.holdParticles()
  model.setCosmoParams(cosmo_par)
  model.forwardModel(delta_hat / box.L[0]**3, final_delta_lpt)
  
  print(f"Num particles = {model.getNumberOfParticles()}")
  Npart = model.getNumberOfParticles()
  print(f"Grid = {Npart**(1./3)}")
  x=np.empty((Npart,3),dtype=np.float64)
  v=np.empty((Npart,3),dtype=np.float64)
  model.getParticlePositions(x)
  model.getParticleVelocities(v)
  
  del model
  
  #
  # Apply PM
  final_delta_pm = np.zeros(outN)
  model = borg.forward.models.BorgPM(box,ai=0.1, nsteps=5, particle_factor=2, supersampling=2)
  model.setAdjointRequired(False)
  model.setCosmoParams(cosmo_par)
  model.forwardModel(delta_hat / box.L[0]**3, final_delta_pm)
  del model

  #
  # Apply PM
  final_delta_pm2 = np.zeros(outN)
  model = borg.forward.models.BorgPM(box,ai=0.1, nsteps=5, particle_factor=2, supersampling=2, force_factor=2)
  model.setAdjointRequired(False)
  model.setCosmoParams(cosmo_par)
  model.forwardModel(delta_hat / box.L[0]**3, final_delta_pm2)
  del model
  
  #
  # Apply PM-RSD
  final_delta_pm_rsd = np.zeros(outN)
  model = borg.forward.models.BorgPM(box,ai=0.1, nsteps=5, particle_factor=2,rsd=True)
  model.setAdjointRequired(False)
  model.setCosmoParams(cosmo_par)
  model.forwardModel(delta_hat / box.L[0]**3, final_delta_pm_rsd)
  del model
  
  np.savez("results.npz", lpt=final_delta_lpt, tlpt=final_delta_2lpt, pm=final_delta_pm, pm2=final_delta_pm2, pm_rsd=final_delta_pm_rsd)



import matplotlib.pyplot as plt
import h5py as h5
import subprocess as sp


models={'lpt':'LPT','old_pm':'PMv1','pm':'PMv2','tcola':'tCOLA'}
#models={'lpt':'LPT','pm':'PMv2'}
output={}
output_v={}
Pref=None
km=None

for m,name in models.items():
  sp.run(f"./libLSS/tests/test_forward_{m}", check=True)
  with h5.File("dump.h5", mode="r") as ff:
      output[m] = ff['power'][:]
      output_v[m] = ff['power_v'][:]
      Pref = ff['scalars/powerspectrum'][:]
      km = ff['scalars/k_modes'][:]

#km = 0.5*(km[1:]+km[:-1])

fig2 = plt.figure(figsize=(5,5))
fig  = plt.figure(figsize=(5,6))
ax0 = plt.axes([0.1 ,0.2,0.9 ,0.7], rasterized=False)
ax1 = plt.axes([0.1 ,0.05,0.9 ,0.15], rasterized=False)
ax2_0 = fig2.add_subplot(111)
ax0.set_xlim(1e-3,2.2)
ax0.set_ylim(1e2,1e5)
ax1.set_xlim(1e-3,2.2)
ax0.set_xticks([])
ax1.set_xlabel('k ($h$ Mpc$^{-1}$)')
ax1.set_ylabel('$P(k) / P_{ref}(k)$')
ax0.set_ylabel('$P(k)$ ($h^{-3}$ Mpc$^{3}$)')
ax2_0.set_ylabel('$P(k)$ (km$^2$ s$^{-2}$ $h^{-3}$ Mpc$^{3}$)')
ax2_0.set_xlabel('k ($h$ Mpc$^{-1}$)')
ax2_0.set_ylim(1, 1e7)


for m,name in models.items():
  ax0.loglog(km,output[m],label=name)
  ax1.semilogx(km, output[m]/Pref,label=name)
  ax2_0.loglog(km, output_v[m], label=name)

ax0.loglog(km,Pref)
ax0.legend()
ax1.set_ylim(0.,1.5)
ax1.axhline(1.0,lw=1.0,color='k')
ax2_0.legend()
fig.savefig("models.pdf",bbox_inches='tight')
fig2.savefig("v_models.pdf",bbox_inches='tight')
